from enum import Enum


class ConditionOperationPattern(Enum):
    EQUAL_TO = r"@equals\((.+),\s*(.+)\)"
    GREATER_THAN = r"@greater\((.+),\s*(.+)\)"
    GREATER_THAN_OR_EQUAL = r"@greaterOrEquals\((.+),\s*(.+)\)"
    LESS_THAN = r"@less\((.+),\s*(.+)\)"
    LESS_THAN_OR_EQUAL = r"@lessOrEquals\((.+),\s*(.+)\)"
    NOT_EQUAL = r"@not\(equals\((.+),\s*(.+)\)\)"



import os

# Get the project root directory (go up from src/wkmigrate/ to the project root)
_project_root = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))

# Path to test JSON files
JSON_PATH = os.path.join(_project_root, "tests", "resources", "json")

# Path to test YAML files
YAML_PATH = os.path.join(_project_root, "tests", "resources", "yaml")



"""This module defines the version information for the wkmigrate library.

Management of version identifiers for releases uses the `bumpversion` python package which
supports automatic modification of files when the version labels need to be modified.

See: https://pypi.org/project/bumpversion/

Note the use of `get_version` for method name to conform with bumpversion conventions.
"""

from collections import namedtuple
import re
import logging


VersionInfo = namedtuple("VersionInfo", ["major", "minor", "patch", "release", "build"])


def get_version(version: str) -> VersionInfo:
    """Gets the version info object for wkmigrate.
    :parameter version: Version to get info for as a ``str``; must be compatible with the `bump` package
    :return: Version info as a ``namedtuple``
    """
    version_expression = re.compile(
        r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+){0,1}(?P<release>\D*)(?P<build>\d*)"
    )
    match = version_expression.match(version)
    if match is None:
        raise ValueError(f"Invalid version format: {version}")
    major, minor, patch, release, build = match.groups()
    version_info = VersionInfo(major, minor, patch, release, build)
    logger = logging.getLogger(__name__)
    logger.info(f"Version : {version_info}")
    return version_info


__version__ = "0.0.0"  # NOTE: MANAGED BY BUMPVERSION; DO NOT EDIT DIRECTLY
__version_info__ = get_version(__version__)



"""This module defines methods for translating Azure SQL Server linked services from data pipeline definitions."""

from wkmigrate.utils import identity, translate


mapping = {
    "host": {"key": "server", "parser": identity},
    "database": {"key": "database", "parser": identity},
    "user_name": {"key": "user_name", "parser": identity},
    "authentication_type": {"key": "authentication_type", "parser": identity},
}


def translate_sql_server_spec(sql_server_spec: dict | None) -> dict | None:
    """Translates an Azure SQL Server linked service in Data Factory's object model to a set of parameters
    used by the Spark JDBC connector for MSSQL.
    :parameter sql_server_spec: Azure SQL Server linked service definition as a ``dict``
    :return: Spark JDBC SQL Server specification as a ``dict``
    """
    if sql_server_spec is None:
        return None
    # Get the cluster properties:
    properties = sql_server_spec.get("properties")
    # Translate the properties:
    translated_properties = translate(properties, mapping)
    result = {"service_name": sql_server_spec.get("name")}
    if translated_properties is not None:
        result.update(translated_properties)
    return result



"""This module defines methods for translating data pipelines."""

import warnings
from wkmigrate.activity_translators.activity_translator import translate_activities
from wkmigrate.pipeline_translators.parameter_translator import translate_parameters
from wkmigrate.trigger_translators.schedule_trigger_translator import (
    translate_schedule_trigger,
)
from wkmigrate.utils import append_system_tags


def translate_pipeline(pipeline: dict) -> dict:
    """Translates a data pipeline to a common object model.
    :parameter pipeline: Dictionary definition of the source pipeline
    :return: Dictionary definition of the target workflows"""
    if "name" not in pipeline:
        warnings.warn(
            "No pipeline name in source definition, setting to UNNAMED_WORKFLOW",
            stacklevel=2,
        )
    # Translate the pipeline:
    translated_pipeline = {
        "name": pipeline.get("name", "UNNAMED_WORKFLOW"),
        "parameters": translate_parameters(pipeline.get("parameters")),
        "schedule": translate_schedule_trigger(pipeline["trigger"]) if pipeline.get("trigger") is not None else None,
        "tasks": translate_activities(pipeline.get("activities")),
        "tags": append_system_tags(pipeline.get("tags")),
    }
    return translated_pipeline



"""This module defines methods for parsing dataset properties to the Databricks SDK's object model."""

import json
from datetime import datetime, timedelta
from wkmigrate.enums.isolation_level import IsolationLevel
from wkmigrate.linked_service_translators.abfs_linked_service_translator import (
    translate_abfs_spec,
)
from wkmigrate.linked_service_translators.sql_server_linked_service_translator import (
    translate_sql_server_spec,
)
from wkmigrate.linked_service_translators.databricks_linked_service_translator import (
    translate_cluster_spec,
)
from wkmigrate.utils import identity, translate


def parse_avro_file_dataset(dataset: dict) -> dict:
    """Parses avro dataset properties to Spark avro reader properties.
    :parameter dataset: Avro dataset properties as a ``dict``
    :return: Spark avro reader properties as a ``dict``
    """
    mapping = {
        "dataset_name": {"key": "name", "parser": identity},
        "container": {"key": "properties", "parser": _parse_abfs_container_name},
        "folder_path": {"key": "properties", "parser": _parse_abfs_file_path},
        "compression_codec": {"key": "avro_compression_codec", "parser": identity},
    }
    translated_dataset = translate(dataset, mapping)
    linked_service_def = dataset.get("linked_service_definition")
    translated_abfs = translate_abfs_spec(linked_service_def)

    result = {}
    if translated_dataset is not None:
        result.update(translated_dataset)
    if translated_abfs is not None:
        result.update(translated_abfs)
    return result


def parse_avro_file_properties(properties: dict) -> dict:
    """Parses avro file properties to Spark avro reader properties.
    :parameter properties: Avro file properties as a ``dict``
    :return: Spark avro reader properties as a ``dict``
    """
    mapping = {
        "type": {"key": "type", "parser": _parse_dataset_type},
        "records_per_file": {
            "key": "format_settings",
            "parser": lambda x: x.get("max_rows_per_file"),
        },
        "file_path_prefix": {
            "key": "format_settings",
            "parser": lambda x: x.get("file_name_prefix"),
        },
    }
    translated = translate(properties, mapping)
    return translated if translated is not None else {}


def parse_delimited_file_dataset(dataset: dict) -> dict:
    """Parses delimited dataset properties to Spark CSV reader properties.
    :parameter dataset: Delimited dataset properties as a ``dict``
    :return: Spark CSV reader properties as a ``dict``
    """
    mapping = {
        "dataset_name": {"key": "name", "parser": identity},
        "container": {"key": "properties", "parser": _parse_abfs_container_name},
        "folder_path": {"key": "properties", "parser": _parse_abfs_file_path},
        "sep": {
            "key": "properties",
            "parser": lambda x: _parse_character_value(x.get("column_delimiter")),
        },
        "lineSep": {
            "key": "properties",
            "parser": lambda x: _parse_character_value(x.get("row_delimiter")),
        },
        "header": {
            "key": "properties",
            "parser": lambda x: x.get("first_row_as_header"),
        },
        "quote": {
            "key": "properties",
            "parser": lambda x: _parse_character_value(x.get("quote_char")),
        },
        "escape": {
            "key": "properties",
            "parser": lambda x: _parse_character_value(x.get("escape_char")),
        },
        "nullValue": {
            "key": "properties",
            "parser": lambda x: _parse_character_value(x.get("null_value")),
        },
        "compression": {
            "key": "properties",
            "parser": lambda x: x.get("compression_codec"),
        },
        "encoding": {"key": "properties", "parser": lambda x: x.get("encoding_name")},
    }
    translated_dataset = translate(dataset, mapping)
    linked_service_def = dataset.get("linked_service_definition")
    translated_abfs = translate_abfs_spec(linked_service_def)

    result = {}
    if translated_dataset is not None:
        result.update(translated_dataset)
    if translated_abfs is not None:
        result.update(translated_abfs)
    return result


def parse_delimited_file_properties(properties: dict) -> dict:
    """Parses delimited file properties to Spark CSV reader properties.
    :parameter properties: Delimited file properties as a ``dict``
    :return: Spark CSV reader properties as a ``dict``
    """
    mapping = {
        "type": {"key": "type", "parser": _parse_dataset_type},
        "quoteAll": {
            "key": "format_settings",
            "parser": lambda x: x.get("quote_all_text"),
        },
        "records_per_file": {
            "key": "format_settings",
            "parser": lambda x: x.get("max_rows_per_file"),
        },
    }
    translated = translate(properties, mapping)
    return translated if translated is not None else {}


def parse_delta_table_dataset(dataset: dict) -> dict:
    """Parses Delta dataset properties to Delta table reader properties.
    :parameter dataset: Delta dataset properties as a ``dict``
    :return: Delta table reader properties as a ``dict``
    """
    mapping = {
        "dataset_name": {"key": "name", "parser": identity},
        "database_name": {"key": "properties", "parser": lambda x: x.get("database")},
        "table_name": {"key": "properties", "parser": lambda x: x.get("table")},
    }
    translated_dataset = translate(dataset, mapping)
    linked_service_def = dataset.get("linked_service_definition")
    translated_cluster = translate_cluster_spec(linked_service_def)

    result = {}
    if translated_dataset is not None:
        result.update(translated_dataset)
    if translated_cluster is not None:
        result.update(translated_cluster)
    return result


def parse_delta_properties(properties: dict) -> dict:
    """Parses Delta file to Delta table reader properties.
    :parameter properties: Delta file properties as a ``dict``
    :return: Delta table reader properties as a ``dict``
    """
    mapping = {"type": {"key": "type", "parser": _parse_dataset_type}}
    translated = translate(properties, mapping)
    return translated if translated is not None else {}


def parse_json_file_dataset(dataset: dict) -> dict:
    """Parses JSON dataset properties to Spark JSON reader properties.
    :parameter dataset: JSON dataset properties as a ``dict``
    :return: Spark JSON reader properties as a ``dict``
    """
    mapping = {
        "dataset_name": {"key": "name", "parser": identity},
        "container": {"key": "properties", "parser": _parse_abfs_container_name},
        "folder_path": {"key": "properties", "parser": _parse_abfs_file_path},
        "encoding": {"key": "properties", "parser": lambda x: x.get("encoding_name")},
        "compression": {
            "key": "properties",
            "parser": lambda x: _parse_compression_type(x.get("compression")),
        },
    }
    translated_dataset = translate(dataset, mapping)
    linked_service_def = dataset.get("linked_service_definition")
    translated_abfs = translate_abfs_spec(linked_service_def)

    result = {}
    if translated_dataset is not None:
        result.update(translated_dataset)
    if translated_abfs is not None:
        result.update(translated_abfs)
    return result


def parse_json_file_properties(properties: dict) -> dict:
    """Parses JSON file properties to Spark JSON reader properties.
    :parameter properties: JSON file properties as a ``dict``
    :return: Spark JSON reader properties as a ``dict``
    """
    mapping = {
        "type": {"key": "type", "parser": _parse_dataset_type},
        "records_per_file": {
            "key": "format_settings",
            "parser": lambda x: x.get("maxRowsPerFile"),
        },
    }
    translated = translate(properties, mapping)
    return translated if translated is not None else {}


def parse_orc_file_dataset(dataset: dict) -> dict:
    """Parses ORC dataset properties to Spark ORC reader properties.
    :parameter dataset: ORC dataset properties as a ``dict``
    :return: Spark ORC reader properties as a ``dict``
    """
    mapping = {
        "dataset_name": {"key": "name", "parser": identity},
        "container": {"key": "properties", "parser": _parse_abfs_container_name},
        "folder_path": {"key": "properties", "parser": _parse_abfs_file_path},
        "compression": {
            "key": "properties",
            "parser": lambda x: x.get("orc_compression_codec"),
        },
    }
    translated_dataset = translate(dataset, mapping)
    linked_service_def = dataset.get("linked_service_definition")
    translated_abfs = translate_abfs_spec(linked_service_def)

    result = {}
    if translated_dataset is not None:
        result.update(translated_dataset)
    if translated_abfs is not None:
        result.update(translated_abfs)
    return result


def parse_orc_file_properties(properties: dict) -> dict:
    """Parses ORC file properties to Spark ORC reader properties.
    :parameter properties: ORC file properties as a ``dict``
    :return: Spark ORC reader properties as a ``dict``
    """
    mapping = {
        "type": {"key": "type", "parser": _parse_dataset_type},
        "file_name_prefix": {
            "key": "format_settings",
            "parser": lambda x: x.get("file_name_prefix"),
        },
        "records_per_file": {
            "key": "format_settings",
            "parser": lambda x: x.get("max_rows_per_file"),
        },
    }
    translated = translate(properties, mapping)
    return translated if translated is not None else {}


def parse_parquet_file_dataset(dataset: dict) -> dict:
    """Parses parquet dataset properties to Spark parquet reader properties.
    :parameter dataset: Parquet dataset properties as a ``dict``
    :return: Spark parquet reader properties as a ``dict``
    """
    mapping = {
        "dataset_name": {"key": "name", "parser": identity},
        "container": {"key": "properties", "parser": _parse_abfs_container_name},
        "folder_path": {"key": "properties", "parser": _parse_abfs_file_path},
        "compression": {
            "key": "properties",
            "parser": lambda x: x.get("compression_codec"),
        },
    }
    translated_dataset = translate(dataset, mapping)
    linked_service_def = dataset.get("linked_service_definition")
    translated_abfs = translate_abfs_spec(linked_service_def)

    result = {}
    if translated_dataset is not None:
        result.update(translated_dataset)
    if translated_abfs is not None:
        result.update(translated_abfs)
    return result


def parse_parquet_file_properties(properties: dict) -> dict:
    """Parses parquet file properties to Spark parquet reader properties.
    :parameter properties: Parquet properties as a ``dict``
    :return: Spark parquet reader properties as a ``dict``
    """
    mapping = {
        "type": {"key": "type", "parser": _parse_dataset_type},
        "file_name_prefix": {
            "key": "format_settings",
            "parser": lambda x: x.get("file_name_prefix"),
        },
        "records_per_file": {
            "key": "format_settings",
            "parser": lambda x: x.get("max_rows_per_file"),
        },
    }
    translated = translate(properties, mapping)
    return translated if translated is not None else {}


def parse_sql_server_dataset(dataset: dict) -> dict:
    """Parses SQL Server dataset properties to Spark SQL Server connector properties.
    :parameter dataset: SQL Server dataset properties as a ``dict``
    :return: Spark SQL Server connector properties as a ``dict``
    """
    mapping = {
        "dataset_name": {"key": "name", "parser": identity},
        "schema_name": {
            "key": "properties",
            "parser": lambda x: x.get("schema_type_properties_schema"),
        },
        "table_name": {"key": "properties", "parser": lambda x: x.get("table")},
    }
    translated_dataset = translate(dataset, mapping)
    linked_service_def = dataset.get("linked_service_definition")
    translated_sql_server = translate_sql_server_spec(linked_service_def)

    result = {}
    if translated_dataset is not None:
        result.update(translated_dataset)
    if translated_sql_server is not None:
        result.update(translated_sql_server)
    return result


def parse_sql_server_properties(properties: dict) -> dict:
    """Parses a set of SQL Server query properties to Spark SQL Server connector properties.
    :parameter properties: SQL Server query properties as a ``dict``
    :return: Spark SQL Server connection properties as a ``dict``
    """
    mapping = {
        "type": {"key": "type", "parser": _parse_dataset_type},
        "query_isolation_level": {
            "key": "isolation_level",
            "parser": _parse_query_isolation_level,
        },
        "query_timeout_seconds": {
            "key": "query_timeout",
            "parser": _parse_query_timeout_seconds,
        },
    }
    translated = translate(properties, mapping)
    return translated if translated is not None else {}


def _parse_character_value(char: str) -> str:
    return json.dumps(char).strip('"')


def _parse_dataset_type(dataset_type: str) -> str:
    """Parses a Data Factory dataset type to a Spark DataFrameReader format.
    :parameter dataset_type: Data Factory dataset type as a ``str``
    :return: Spark DataFrameReader format as a ``str``
    """
    mappings = {
        "AvroSource": "avro",
        "AvroSink": "avro",
        "AzureDatabricksDeltaLakeSource": "delta",
        "AzureDatabricksDeltaLakeSink": "delta",
        "AzureSqlSource": "sqlserver",
        "AzureSqlSink": "sqlserver",
        "DelimitedTextSource": "csv",
        "DelimitedTextSink": "csv",
        "JsonSource": "json",
        "JsonSink": "json",
        "OrcSource": "orc",
        "OrcSink": "orc",
        "ParquetSource": "parquet",
        "ParquetSink": "parquet",
    }
    result = mappings.get(dataset_type)
    if result is None:
        raise ValueError(f"Unsupported dataset type: {dataset_type}")
    return result


def _parse_compression_type(compression: dict) -> str | None:
    return compression.get("type")


def _parse_query_timeout_seconds(properties: dict | None) -> int:
    """Parses the timeout number of seconds from the dataset properties.
    :parameter properties: Dataset properties as a ``dict``
    :return: Timeout seconds as an ``int``
    """
    if properties is None or "query_timeout" not in properties:
        return 0
    query_timeout = properties.get("query_timeout")
    if query_timeout is None:
        return 0
    return _parse_query_timeout_string(query_timeout)


def _parse_query_isolation_level(properties: dict | None) -> str | None:
    """Parses the database transaction isolation level from the dataset properties.
    :parameter properties: Dataset properties as a ``dict``
    :return: Isolation level as an ``str``
    """
    if properties is None or "isolation_level" not in properties:
        return "READ_COMMITTED"
    isolation_level = properties.get("isolation_level")
    if isolation_level is None:
        return "READ_COMMITTED"
    return IsolationLevel[isolation_level]


def _parse_query_timeout_string(timeout_string: str) -> int:
    """Parses a timeout string in the format ``hh:mm:ss`` into an integer number of seconds.
    :parameter timeout_string: Timeout string in the format ``hh:mm:ss``
    :return: Integer number of seconds
    """
    time_format = "%H:%M:%S"
    date_time = datetime.strptime(timeout_string, time_format)
    time_delta = timedelta(hours=date_time.hour, minutes=date_time.minute, seconds=date_time.second)
    return int(time_delta.total_seconds())


def _parse_abfs_container_name(properties: dict) -> str:
    """Parses an ABFS container name from the full ABFS location in a file properties object.
    :parameter properties: File properties as a ``dict``
    :return: The ABFS container name as a ``str``
    """
    location = properties.get("location")
    if location is None:
        raise ValueError("Location cannot be None")
    return location.get("file_system")


def _parse_abfs_file_path(properties: dict) -> str:
    """Parses an ABFS file path from the full ABFS location in a file properties object.
    :parameter properties: File properties as a ``dict``
    :return: The ABFS file path as a ``str``
    """
    location = properties.get("location")
    if location is None:
        raise ValueError("Location cannot be None")
    return location.get("folder_path")



"""This module defines methods for parsing trigger objects to the Databricks SDK's object model."""

import warnings
from wkmigrate.enums.interval_type import IntervalType


def parse_cron_expression(recurrence: dict | None) -> str | None:
    """Generates a cron expression from a set of schedule trigger parameters.
    :parameter recurrence: Recurrence object as a ``IntervalType``
    :return: Cron expression for the specified parameters as a ``str``
    """
    if recurrence is None:
        return None
    interval_type = recurrence.get("frequency")
    num_intervals = recurrence.get("interval")
    if num_intervals is None:
        warnings.warn('Setting empty "num_intervals" to "1" by default', stacklevel=2)
        num_intervals = 1
    schedule = recurrence.get("schedule")
    if interval_type == IntervalType.HOUR:
        return _get_hourly_cron_expression(num_intervals)
    if interval_type == IntervalType.DAY:
        if schedule is None:
            return _get_daily_cron_expression(num_intervals, {})
        return _get_daily_cron_expression(num_intervals, schedule)
    if interval_type == IntervalType.WEEK:
        if num_intervals > 1:
            warnings.warn(
                'Ignoring "num_intervals" > 1 for weekly triggers; Using weekly interval',
                stacklevel=2,
            )
        if schedule is None:
            return _get_weekly_cron_expression({})
        return _get_weekly_cron_expression(schedule)
    if interval_type == IntervalType.MONTH:
        if num_intervals > 1:
            warnings.warn(
                'Ignoring "num_intervals" > 1 for monthly triggers; Using monthly interval',
                stacklevel=2,
            )
        if schedule is None:
            return _get_monthly_cron_expression({})
        return _get_monthly_cron_expression(schedule)
    return None


def _get_hourly_cron_expression(num_intervals: int) -> str:
    return f"0 0 */{num_intervals} * * ?"


def _get_daily_cron_expression(num_intervals: int, schedule: dict) -> str:
    if schedule is None:
        return f"0 0 0 */{num_intervals} * ?"
    minutes = ",".join([str(e) for e in schedule.get("minutes", [0])])
    hours = ",".join([str(e) for e in schedule.get("hours", [0])])
    return f"0 {minutes} {hours} */{num_intervals} * ?"


def _get_weekly_cron_expression(schedule: dict) -> str:
    if schedule is None:
        return "0 0 0 ? * 1"
    minutes = ",".join([str(e) for e in schedule.get("minutes", [0])])
    hours = ",".join([str(e) for e in schedule.get("hours", [0])])
    week_days = ",".join([_get_week_day(e) for e in schedule.get("week_days", ["Sunday"])])
    return f"0 {minutes} {hours} ? * {week_days}"


def _get_monthly_cron_expression(schedule: dict) -> str:
    if schedule is None:
        return "0 0 0 0 * ?"
    minutes = ",".join([str(e) for e in schedule.get("minutes", [0])])
    hours = ",".join([str(e) for e in schedule.get("hours", [0])])
    days = ",".join([str(e) for e in schedule.get("days", [0])])
    return f"0 {minutes} {hours} {days} * ?"


def _get_week_day(week_day: str) -> str:
    if week_day == "Sunday":
        return "1"
    if week_day == "Monday":
        return "2"
    if week_day == "Tuesday":
        return "3"
    if week_day == "Wednesday":
        return "4"
    if week_day == "Thursday":
        return "5"
    if week_day == "Friday":
        return "6"
    if week_day == "Saturday":
        return "7"
    raise ValueError('Invalid value for parameter "week_day"')



"""This module defines methods for translating Databricks parameter values from data pipelines."""

from wkmigrate.pipeline_translators.parsers import parse_parameter_value
from wkmigrate.utils import translate


mapping = {"default": {"key": "default_value", "parser": parse_parameter_value}}


def translate_parameters(parameters: dict | None) -> list[dict] | None:
    """Translates a set of parameter definitions in the Data Factory object model to the Databricks SDK object model.
    :parameter parameters: List of parameter definitions as a nested ``dict``
    :return: List of translated parameter definitions as ``dict`` objects
    """
    if parameters is None:
        return None
    return [translate_parameter(parameter_name, parameter_def) for parameter_name, parameter_def in parameters.items()]


def translate_parameter(parameter_name: str, parameter: dict) -> dict:
    """Translates a parameter definition in the Data Factory object model to the Databricks SDK object model.
    :parameter parameter_name: Parameter name as a ``str``
    :parameter parameter: Parameter definition as a ``dict``
    :return: Translated parameter definition as a ``dict``
    """
    translated_parameter = translate(parameter, mapping)
    result = {"name": parameter_name}
    if translated_parameter is not None:
        result.update(translated_parameter)
    return result



"""This module defines methods for translating For Each activities."""

from wkmigrate.activity_translators.parsers import (
    parse_for_each_items,
    parse_for_each_tasks,
)
from wkmigrate.utils import identity, translate


mapping = {
    "inputs": {"key": "items", "parser": parse_for_each_items},
    "concurrency": {"key": "batch_count", "parser": identity},
    "task": {"key": "activities", "parser": parse_for_each_tasks},
}


def translate_for_each_activity(activity: dict) -> dict:
    """Translates a For Each activity definition in Data Factory's object model to a Databricks for-each condition
    task in the Databricks SDK object model.
    :parameter activity: For Each activity definition as a ``dict``
    :return: Databricks for-each condition task properties as a ``dict``
    """
    translated = translate(activity, mapping)
    if translated is None:
        raise ValueError('Translation failed')
    return translated



"""This module defines the ``FactoryDefinitionStore`` class."""

import warnings
from dataclasses import dataclass, field
from collections.abc import Callable
from wkmigrate.clients.factory_client import (
    FactoryClient,
    FactoryManagementClient,
    FactoryTestClient,
)
from wkmigrate.definition_stores.definition_store import DefinitionStore
from wkmigrate.pipeline_translators.pipeline_translator import translate_pipeline


@dataclass
class FactoryDefinitionStore(DefinitionStore):
    """This class is used to list and describe objects in an Azure Data Factory instance."""

    tenant_id: str | None = None
    client_id: str | None = None
    client_secret: str | None = None
    subscription_id: str | None = None
    resource_group_name: str | None = None
    factory_name: str | None = None
    factory_client: FactoryClient | None = field(init=False)
    _appenders: list[Callable[[dict], dict]] | None = field(init=False)
    _use_test_client: bool | None = False

    def __post_init__(self) -> None:
        self._appenders = [self._append_datasets, self._append_linked_service]
        if self._use_test_client:
            self.factory_client = FactoryTestClient()
            return
        if self.tenant_id is None:
            raise ValueError("A tenant_id must be provided when creating a FactoryDefinitionStore")
        if self.client_id is None:
            raise ValueError("A client_id must be provided when creating a FactoryDefinitionStore")
        if self.client_secret is None:
            raise ValueError("A client_secret must be provided when creating a FactoryDefinitionStore")
        if self.subscription_id is None:
            raise ValueError("subscription_id cannot be None")
        if self.resource_group_name is None:
            raise ValueError("resource_group_name cannot be None")
        if self.factory_name is None:
            raise ValueError("factory_name cannot be None")

        self.factory_client = FactoryManagementClient(
            tenant_id=self.tenant_id,
            client_id=self.client_id,
            client_secret=self.client_secret,
            subscription_id=self.subscription_id,
            resource_group_name=self.resource_group_name,
            factory_name=self.factory_name,
        )

    def load(self, pipeline_name: str) -> dict:
        """Gets a dictionary representation of a Data Factory pipeline.
        :parameter pipeline_name: Name of the Data Factory pipeline as an ``str``
        :return: Data Factory pipeline as a ``dict``
        """
        if self.factory_client is None:
            raise ValueError("factory_client is not initialized")
        pipeline = self.factory_client.get_pipeline(pipeline_name)
        pipeline["trigger"] = self.factory_client.get_trigger(pipeline_name)
        activities = pipeline.get("activities")
        if activities is not None:
            pipeline["activities"] = [self._append_objects(activity) for activity in activities]
        else:
            pipeline["activities"] = []
        return translate_pipeline(pipeline)

    def dump(self, pipeline_definition: dict) -> None:
        """NOTE: Loading data pipeline definitions to Data Factory is not currently supported."""
        raise NotImplementedError(f"Dump of pipeline {pipeline_definition} to FactoryDefinitionStore not supported.")

    def _append_objects(self, activity: dict) -> dict:
        """Appends datasets, linked services, etc. associated with a pipeline activity.
        :parameter activity: Data Factory activity definition as a ``dict``
        :return: Data Factory activity definition with parsed objects as a ``dict``
        """
        if self._appenders is None:
            return activity
        for appender in self._appenders:
            activity = appender(activity)
        return activity

    def _append_datasets(self, activity: dict) -> dict:
        """Gets the dataset definitions and associated linked services for the
        specified pipeline activity.
        :parameter activity: Data Factory activity definition as a ``dict``
        :return: Data Factory activity definition with parsed datasets and linked services as a ``dict``
        """
        if "inputs" in activity:
            datasets = activity.get("inputs")
            if datasets is not None:
                dataset_names = [dataset.get("reference_name") for dataset in datasets]
                if self.factory_client is None:
                    raise ValueError("factory_client is not initialized")
                activity["input_dataset_definitions"] = [
                    self.factory_client.get_dataset(dataset_name) for dataset_name in dataset_names
                ]
        if "outputs" in activity:
            datasets = activity.get("outputs")
            if datasets is not None:
                dataset_names = [dataset.get("reference_name") for dataset in datasets]
                if self.factory_client is None:
                    raise ValueError("factory_client is not initialized")
                activity["output_dataset_definitions"] = [
                    self.factory_client.get_dataset(dataset_name) for dataset_name in dataset_names
                ]
        return activity

    def _append_linked_service(self, activity: dict) -> dict:
        """Gets the Databricks linked service for the specified pipeline activity
        :parameter activity: Data Factory activity definition as a ``dict``
        :return: Data Factory activity definition with parsed linked service as a ``dict``
        """
        if "linked_service_name" in activity:
            # Get the linked service reference name:
            linked_service_reference = activity.get("linked_service_name")
            if linked_service_reference is not None:
                linked_service_name = linked_service_reference.get("reference_name")
                if self.factory_client is None:
                    raise ValueError("factory_client is not initialized")
                # Get the linked service details from data factory:
                linked_service = self.factory_client.get_linked_service(linked_service_name)
                if linked_service["type"] == "AzureDatabricks":
                    activity["linked_service_definition"] = self.factory_client.get_linked_service(linked_service_name)

        # Check the nested "if false" activities:
        if "if_false_activities" in activity:
            if_false_activities = activity.get("if_false_activities")
            if if_false_activities is not None:
                activity["if_false_activities"] = [
                    self._append_linked_service(if_false_activity) for if_false_activity in if_false_activities
                ]
        # Check the nested "if true" activities:
        if "if_true_activities" in activity:
            if_true_activities = activity.get("if_true_activities")
            if if_true_activities is not None:
                activity["if_true_activities"] = [
                    self._append_linked_service(if_true_activity) for if_true_activity in if_true_activities
                ]
        # Check the nested "for each" activities:
        if "activities" in activity:
            activities = activity.get("activities")
            if activities is not None:
                activity["activities"] = [
                    self._append_linked_service(for_each_activity) for for_each_activity in activities
                ]
        return activity



"""This module defines methods for translating activities from data pipelines."""

from wkmigrate.activity_translators import type_mapping, type_translators
from wkmigrate.activity_translators.parsers import parse_dependencies, parse_policy
from wkmigrate.linked_service_translators.databricks_linked_service_translator import (
    translate_cluster_spec,
)
from wkmigrate.utils import identity, translate


mapping = {
    "type": {"key": "type", "parser": identity},
    "task_key": {
        "key": "name",
        "parser": lambda x: x if x else "TASK_NAME_NOT_PROVIDED",
    },
    "description": {"key": "description", "parser": identity},
    "timeout_seconds": {
        "key": "policy",
        "parser": lambda x: parse_policy(x).get("timeout_seconds"),
    },
    "max_retries": {
        "key": "policy",
        "parser": lambda x: parse_policy(x).get("max_retries"),
    },
    "min_retry_interval_millis": {
        "key": "policy",
        "parser": lambda x: parse_policy(x).get("min_retry_interval_millis"),
    },
    "depends_on": {"key": "depends_on", "parser": parse_dependencies},
    "new_cluster": {
        "key": "linked_service_definition",
        "parser": translate_cluster_spec,
    },
}


def translate_activities(activities: list[dict] | None) -> list[dict] | None:
    """Translates a set of data factory pipeline activities to a common object model.
    :parameter activities: Source pipeline activities
    :return: Target pipeline activities
    """
    if activities is None:
        return None
    translated_activities = []
    for activity in activities:
        translated_activity = translate_activity(activity)
        if isinstance(translated_activity, tuple):
            translated_activities.append(translated_activity[0])
            child_activities = translate_activities(translated_activity[1])
            if child_activities is not None:
                translated_activities.extend(child_activities)
            continue
        translated_activities.append(translated_activity)
    return translated_activities


def translate_activity(activity: dict) -> dict | tuple[dict, list[dict]]:
    """Translates a data pipeline activity to a common object model.
    :parameter activity: Dictionary definition of the source pipeline activity
    :return: Dictionary definition of the target workflows activity"""
    # Translate the activity properties:
    translated_activity = translate(activity, mapping)
    if translated_activity is None:
        translated_activity = {}
    # Parse the type properties for the task type:
    parsed_properties = parse_activity_properties(activity)
    # Check if any downstream activities were created:
    if isinstance(parsed_properties, tuple):
        if parsed_properties[0] is not None:
            return ({**translated_activity, **parsed_properties[0]}), parsed_properties[1]
        return translated_activity, parsed_properties[1]
    if parsed_properties is not None:
        return {**translated_activity, **parsed_properties}
    return translated_activity


def parse_activity_properties(activity: dict) -> dict | tuple[dict, list[dict]]:
    """Parses a data factory pipeline activity policy to a common object model.
    :parameter activity: Pipeline activity in the data factory object model as ``dict``
    :return: Workflow task in the Databricks SDK object model as ``dict``
    """
    translated_activity = {}
    # Translate the activity:
    activity_type = activity.get("type")
    if activity_type is None:
        raise ValueError("Activity type cannot be None")

    type_translator = type_translators.get(activity_type)
    if not type_translator:
        return get_placeholder_activity()

    translated_task = type_translator(activity)
    # Check if any downstream activities were created:
    if isinstance(translated_task, tuple):
        translated_activity[type_mapping[activity_type]] = translated_task[0]
        return translated_activity, translated_task[1]
    translated_activity[type_mapping[activity_type]] = translated_task
    return translated_activity


def get_placeholder_activity() -> dict:
    return {"notebook_task": {"notebook_path": "/UNSUPPORTED_ADF_ACTIVITY"}, "unsupported": True}



"""This module defines the ``DatabricksWorkspaceDefinitionStore`` class."""

from dataclasses import dataclass, field
from wkmigrate.clients.workspace_client import (
    DatabricksWorkspaceClient,
    WorkspaceManagementClient,
    WorkspaceTestClient,
)
from wkmigrate.definition_stores.definition_store import DefinitionStore


@dataclass
class WorkspaceDefinitionStore(DefinitionStore):
    """This class is used to list, describe, and update objects in a Databricks workspace."""

    authentication_type: str | None = None
    host_name: str | None = None
    pat: str | None = None
    username: str | None = None
    password: str | None = None
    resource_id: str | None = None
    tenant_id: str | None = None
    client_id: str | None = None
    client_secret: str | None = None
    files_to_delta_sinks: bool | None = True
    workspace_client: DatabricksWorkspaceClient | None = field(init=False)
    _use_test_client: bool | None = False
    _valid_authentication_types = ["pat", "basic", "azure-client-secret"]

    def __post_init__(self) -> None:
        """Sets up the workspace client for the provided authentication credentials."""
        if self._use_test_client:
            self.workspace_client = WorkspaceTestClient()
            return
        if self.authentication_type not in self._valid_authentication_types:
            raise ValueError(
                'Invalid value for "self.authentication_type"; Must be "pat", "basic", or "azure-client-secret"'
            )
        if self.host_name is None:
            raise ValueError('"host_name" must be provided when creating a WorkspaceDefinitionStore')
        self.workspace_client = WorkspaceManagementClient(
            self.authentication_type,
            self.host_name,
            self.pat,
            self.username,
            self.password,
            self.resource_id,
            self.tenant_id,
            self.client_id,
            self.client_secret,
        )

    def load(self, job_name: str) -> dict:
        """Gets a dictionary representation of a Databricks workflow from the Databricks workspace.
        :parameter job_name: Job name for the specified workflow
        :return: Workflow definition as a ``dict``
        """
        if self.workspace_client is None:
            raise ValueError("workspace_client is not initialized")
        job = self.workspace_client.get_workflow(job_name=job_name)
        return job.as_dict()

    def dump(self, job_settings: dict) -> int:
        """Creates workflow in the Databricks workspace with the specified definition.
        :parameter job_settings: Workflow definition as a ``dict``
        :return: ``None``
        """
        job_definition = {"settings": job_settings}
        if self.workspace_client is None:
            raise ValueError("workspace_client is not initialized")
        result = self.workspace_client.create_workflow(job_definition=job_definition)
        if result is None:
            raise ValueError("Failed to create workflow")
        return result



"""This module defines methods for translating Databricks Notebook activities."""

from wkmigrate.activity_translators.parsers import parse_notebook_parameters
from wkmigrate.utils import identity, translate

mapping = {
    "notebook_path": {"key": "notebook_path", "parser": identity},
    "base_parameters": {"key": "base_parameters", "parser": parse_notebook_parameters},
}


def translate_notebook_activity(activity: dict) -> dict:
    """Translates a Databricks notebook activity definition in Data Factory's object model to a Databricks notebook
    task in the Databricks SDK object model.
    :parameter activity: Databricks notebook activity definition as a ``dict``
    :return: Databricks notebook task properties as a ``dict``
    """
    translated = translate(activity, mapping)
    if translated is None:
        raise ValueError('Translation failed')
    return translated



"""This module defines methods for translating Databricks cluster services from data pipeline definitions."""

from wkmigrate.linked_service_translators.parsers import (
    parse_autoscale_policy,
    parse_init_scripts,
    parse_log_conf,
    parse_number_of_workers,
)
from wkmigrate.utils import append_system_tags, identity, translate


mapping = {
    "host_name": {"key": "domain", "parser": identity},
    "node_type_id": {"key": "new_cluster_node_type", "parser": identity},
    "spark_version": {"key": "new_cluster_version", "parser": identity},
    "custom_tags": {"key": "new_cluster_custom_tags", "parser": append_system_tags},
    "driver_node_type_id": {"key": "new_cluster_driver_node_type", "parser": identity},
    "spark_conf": {"key": "new_cluster_spark_conf", "parser": identity},
    "spark_env_vars": {"key": "new_cluster_spark_env_vars", "parser": identity},
    "init_scripts": {"key": "new_cluster_init_scripts", "parser": parse_init_scripts},
    "cluster_log_conf": {
        "key": "new_cluster_log_destination",
        "parser": parse_log_conf,
    },
    "autoscale": {"key": "new_cluster_num_of_worker", "parser": parse_autoscale_policy},
    "num_workers": {
        "key": "new_cluster_num_of_worker",
        "parser": parse_number_of_workers,
    },
    "pat": {"key": "pat", "parser": identity},
}


def translate_cluster_spec(cluster_spec: dict | None) -> dict | None:
    """Translates a Databricks linked service in Data Factory's object model to a Databricks cluster
    specification in the Databricks SDK object model.
    :parameter cluster_spec: Databricks linked service definition as a ``dict``
    :return: Databricks cluster specification as a ``dict``
    """
    if cluster_spec is None:
        return None
    # Get the cluster properties:
    properties = cluster_spec.get("properties")
    # Translate the properties:
    translated_properties = translate(properties, mapping)
    result = {"service_name": cluster_spec.get("name")}
    if translated_properties is not None:
        result.update(translated_properties)
    return result



"""This module defines the abstract ``DefinitionStore`` class."""

from abc import ABC, abstractmethod
from typing import Any


class DefinitionStore(ABC):
    """A source or sink for data pipeline definitions. Used to get or set
    data pipelines during migration.
    """

    @abstractmethod
    def load(self, *args: Any, **kwargs: Any) -> dict:
        """Loads the data pipeline into a dictionary object."""

    @abstractmethod
    def dump(self, *args: Any, **kwargs: Any) -> int | None:
        """Pushes the data pipeline definition into the definition store. This could
        create a file (e.g. JSON or YAML) or a Workflow definition in a Databricks
        workspace."""



"""This module defines methods for translating Databricks Spark jar activities."""

from wkmigrate.utils import identity, translate


mapping = {
    "main_class_name": {"key": "main_class_name", "parser": identity},
    "parameters": {"key": "parameters", "parser": identity},
}


def translate_spark_jar_activity(activity: dict) -> dict:
    """Translates a Databricks Spark jar activity definition in Data Factory's object model to a Databricks Spark jar
    task in the Databricks SDK object model.
    :parameter activity: Databricks Spark jar activity definition as a ``dict``
    :return: Databricks Spark jar task properties as a ``dict``
    """
    translated = translate(activity, mapping)
    if translated is None:
        raise ValueError('Translation failed')
    return translated



"""This module defines methods for parsing cluster spec objects to the Databricks SDK's object model."""

import re
from wkmigrate.enums.init_script_type import InitScriptType


def parse_log_conf(cluster_log_destination: str | None) -> dict | None:
    """Parses a cluster log configuration from a path to DBFS.
    :parameter: cluster_log_destination: Cluster log delivery path in DBFS as an ``str``
    :return: Cluster log configuration as a ``dict``
    """
    if cluster_log_destination is None:
        return None
    return {"dbfs": {"destination": cluster_log_destination}}


def parse_number_of_workers(num_workers: str | None) -> int | None:
    """Parses a cluster number of workers from a string (e.g. '4').
    :parameter num_workers: Number of workers as a ``str``
    :return: Number of workers as an ``int``
    """
    if num_workers is None or ":" in num_workers:
        return None
    return int(num_workers)


def parse_autoscale_policy(autoscale_policy: str | None) -> dict | None:
    """Parses a cluster autoscale policy from a string (e.g. '1:4').
    :parameter autoscale_policy: Autoscale policy as a ``str``
    :return: Min and max worker counts as a ``tuple[int, int]``
    """
    if autoscale_policy is None or ":" not in autoscale_policy:
        return None
    autoscale_num_workers = autoscale_policy.split(":")
    return {
        "min_workers": int(autoscale_num_workers[0]),
        "max_workers": int(autoscale_num_workers[1]),
    }


def parse_init_scripts(init_scripts: list[str] | None) -> list[dict] | None:
    """Parses a list of cluster init script definitions to the Databricks SDK's object model.
    :parameter init_scripts: List of init script paths
    :return: List of init script objects with ``type`` and ``destination`` properties
    """
    if init_scripts is None or len(init_scripts) == 0:
        return None
    return [
        {_get_init_script_type(init_script_path=init_script): {"destination": init_script}}
        for init_script in init_scripts
    ]


def parse_storage_account_name(url: str) -> str:
    """Parses an Azure Storage account name from the URL.
    :parameter url: Azure Storage account URL as a ``str``
    :return: Azure Storage account name as a ``str``
    """
    match = re.search(pattern=r"https://([A-Za-z0-9]*).dfs.core.windows.net", string=url)
    if match is None:
        raise ValueError(f"Invalid Azure Storage URL format: {url}")
    return match[1]


def _get_init_script_type(init_script_path: str) -> str:
    """Gets an init script type from the path.
    :parameter: init_script_path: Init script path as a ``str``
    :return: Init script type as a ``str``
    """
    if init_script_path.startswith("dbfs:"):
        return InitScriptType.DBFS.value
    if init_script_path.startswith("/Volumes"):
        return InitScriptType.VOLUMES.value
    return InitScriptType.WORKSPACE.value



import ast
import re
import warnings
from datetime import datetime, timedelta
from importlib import import_module
from wkmigrate.datasets import dataset_parsers, property_parsers
from wkmigrate.enums.condition_operation_pattern import ConditionOperationPattern


def parse_dataset(datasets: list[dict]) -> dict:
    dataset = datasets[0]
    properties = dataset.get("properties")
    if properties is None:
        raise ValueError("Dataset properties cannot be None")
    dataset_type = properties.get("type")
    if dataset_type is None:
        raise ValueError("Dataset type cannot be None")
    return dataset_parsers[dataset_type](dataset)


def parse_dataset_mapping(mapping: dict) -> list[dict]:
    """Parses a mapping from one set of data columns to another.
    :parameter mapping: Data column mapping as a ``dict``
    :return: Parsed data column mapping as a ``list[dict]``
    """
    return [
        {
            "source_column_name": mapping.get("source").get("name"),
            "sink_column_name": mapping.get("sink").get("name"),
            "sink_column_type": mapping.get("sink").get("type"),
        }
        for mapping in (mapping.get("mappings") or [])
    ]


def parse_dataset_properties(dataset_definition: dict) -> dict:
    """Parses various properties (e.g. query timeout, isolation level) from an input dataset definition.
    :parameter dataset_definition: Dataset definition as a ``dict``
    :return: Parsed dataset properties as a ``dict``
    """
    dataset_type = dataset_definition.get("type")
    if dataset_type is None:
        raise ValueError("Dataset type cannot be None")
    return property_parsers[dataset_type](dataset_definition)


def parse_for_each_tasks(tasks: list[dict] | None) -> list[dict] | None:
    """Parses multiple task definitions within a ForEach task to a common object model.
    :parameter tasks: List of source pipeline ForEach task definitions as ``dict`` objects
    :return: List of common object model ForEach task definitions as ``dict`` objects
    """
    if tasks is None:
        return None
    parsed_tasks = [_parse_for_each_task(task) for task in tasks]
    return [task for task in parsed_tasks if task is not None]


def parse_for_each_items(items: dict | None) -> str | None:
    """Parses a list of items passed to a ForEach task to a common object model.
    :parameter items: Set of values passed to a ForEach activity as a ``str``.
    :return: List of values to pass to the for-each task
    """
    if items is None:
        return None
    if "value" not in items:
        raise ValueError('For Each task must specify "value" property in "items" list')
    value = items.get("value")
    if value is None:
        return None
    # TODO: Move all dynamic function patterns to a common enum list
    array_pattern = r"@array\('(.+)'\)"
    match = re.match(string=value, pattern=array_pattern)
    if match:
        matched_item = match.group(1)
        return f'["{matched_item}"]'

    create_array_pattern = r"@createArray\((.+)\)"
    match = re.match(string=value, pattern=create_array_pattern)
    if match:
        matched_item = match.group(1)
        list_items = ast.literal_eval(matched_item)
        return f'[{",".join([f'"{item}"' for item in list_items])}]'
    return None


def parse_policy(policy: dict | None) -> dict:
    """Parses a data factory pipeline activity policy to a common object model.
    :parameter policy: Dictionary definition of the source pipeline activity policy
    :return: Dictionary definition of the policy settings
    """
    if policy is None:
        return {}
    # Warn about secure input/output logging:
    if "secure_input" in policy:
        warnings.warn("Secure input logging not applicable to Databricks workflows.", stacklevel=2)
    if "secure_output" in policy:
        warnings.warn(
            "Secure output logging not applicable to Databricks workflows.",
            stacklevel=2,
        )
    # Parse the policy attributes:
    parsed_policy = {}
    # Parse the timeout seconds:
    if "timeout" in policy:
        timeout_value = policy.get("timeout")
        if timeout_value is not None:
            parsed_policy["timeout_seconds"] = _parse_activity_timeout_string(timeout_value)
    # Parse the number of retry attempts:
    if "retry" in policy:
        retry_value = policy.get("retry")
        if retry_value is not None:
            parsed_policy["max_retries"] = int(retry_value)
    # Parse the retry wait time in milliseconds:
    if "retry_interval_in_seconds" in policy:
        parsed_policy["min_retry_interval_millis"] = 1000 * int(policy.get("retry_interval_in_seconds", 0))
    return parsed_policy


def parse_dependencies(dependencies: list[dict] | None) -> list[dict] | None:
    """Parses a data factory pipeline activity's dependencies to a common object model.
    :parameter dependencies: Dictionary definition of the source pipeline activity's dependencies
    :return: Dictionary definition of the task-level parameter definitions
    """
    if dependencies is None:
        return None
    # Parse the dependencies from the list:
    parsed_dependencies = []
    for dependency in dependencies:
        # Get the dependency condition:
        conditions = dependency.get("dependencyConditions")
        # Validate the dependency conditions:
        if conditions is not None and len(conditions) > 1:
            raise ValueError("Dependencies with multiple conditions are not supported.")
        # Append the dependency:
        parsed_dependencies.append(
            {
                "task_key": dependency.get("activity", None),
                "outcome": dependency.get("outcome", None),
            }
        )
    return parsed_dependencies


def parse_notebook_parameters(parameters: dict | None) -> dict | None:
    """Parses task parameters in a Databricks notebook activity definition from Data Factory's object model to a
    set of key/value pairs in the Databricks SDK object model.
    :parameter parameters: Set of task parameters in Data Factory as a ``dict[str, Any]``
    :return: Set of task parameters as a ``dict[str, str]``
    """
    if parameters is None:
        return None
    # Parse the parameters:
    parsed_parameters = {}
    for name, value in parameters.items():
        if not isinstance(value, str):
            warnings.warn(
                f'Could not resolve default value for parameter {name}, setting to ""',
                stacklevel=2,
            )
            value = ""
        parsed_parameters[name] = value
    return parsed_parameters


def parse_condition_expression(condition: dict) -> dict:
    """Parses a condition expression in an If Condition activity definition from Data Factory's object model to the
    Databricks SDK object model.
    :parameter condition: Condition expression in Data Factory as a ``dict[str, Any]``
    :return: Condition expression as a ``dict[str, str]``
    """
    # Validate the condition:
    if "value" not in condition:
        raise ValueError("Condition expression must include a valid conditional value")
    # Match a boolean operator:
    condition_value = condition.get("value")
    if condition_value is None:
        raise ValueError("Condition value cannot be None")
    for op in ConditionOperationPattern:
        match = re.match(string=condition_value, pattern=op.value)
        if match is not None:
            return {
                "op": op.name,
                "left": match.group(1).replace('"', "").replace("'", ""),
                "right": match.group(2).replace('"', "").replace("'", ""),
            }
    raise ValueError(
        'Condition expression must include "equals", "greaterThan", "greaterThanOrEquals", "lessThan", or '
        '"lessThanOrEquals" operation.'
    )


def _parse_activity_timeout_string(timeout_string: str) -> int:
    """Parses a timeout string in the format ``d.hh:mm:ss`` into an integer number of seconds.
    :parameter timeout_string: Timeout string in the format ``d.hh:mm:ss``
    :return: Integer number of seconds
    """
    if timeout_string[:2] == "0.":
        # Parse the timeout string to HH:MM:SS format:
        timeout_string = timeout_string[2:]
        time_format = "%H:%M:%S"
        date_time = datetime.strptime(timeout_string, time_format)
        time_delta = timedelta(hours=date_time.hour, minutes=date_time.minute, seconds=date_time.second)
    else:
        # Parse the timeout string to DD.HH:MM:SS format:
        timeout_string = timeout_string.zfill(11)
        time_format = "%d.%H:%M:%S"
        date_time = datetime.strptime(timeout_string, time_format)
        time_delta = timedelta(
            days=date_time.day,
            hours=date_time.hour,
            minutes=date_time.minute,
            seconds=date_time.second,
        )
    return int(time_delta.total_seconds())


def _parse_array_string(array_string: str) -> str:
    """Parses an array string into a JSON-safe format.
    :parameter array_string: Array values as a ``str``
    :return: Array values as a JSON-safe ``str``
    """
    double_quote_character = '"'
    single_quote_character = "'"
    test = f"""["{'","'.join([f'{element.replace(single_quote_character, "").replace(double_quote_character, "")}' for element in array_string.split(',')])}"]"""
    print(test)
    return test

def _parse_for_each_task(task: dict | None) -> dict | None:
    """Parses a single task definition within a ForEach task to a common object model.
    :parameter task: Source pipeline definition of the ForEach task as a ``dict``
    :return: Common object model for each task definition as a ``dict``
    """
    task_with_filtered_parameters = _filter_parameters(task)
    translator = import_module("wkmigrate.activity_translators.activity_translator")
    return getattr(translator, "translate_activity")(task_with_filtered_parameters)


def _filter_parameters(activity: dict | None) -> dict | None:
    if activity is None:
        return None
    if "base_parameters" not in activity:
        warnings.warn("No baseParameters for ForEach inner activity", stacklevel=2)
        return activity
    base_parameters = activity.get("base_parameters")
    if base_parameters is None:
        return activity
    parameters = _filter_parameters(base_parameters)
    if parameters is None:
        return None
    filtered_parameters = {}
    for name, expression in parameters.items():
        if expression is not None and expression.get("value") == "@item()":
            warnings.warn(f"Removing redundant parameter {name} with value {expression.get('value')}")
            continue
        filtered_parameters.update({name: expression})
    activity["base_parameters"] = filtered_parameters
    return activity



"""This module defines methods for translating Azure SQL Server linked services from data pipeline definitions."""

from wkmigrate.linked_service_translators.parsers import parse_storage_account_name
from wkmigrate.utils import identity, translate


mapping = {
    "url": {"key": "url", "parser": identity},
    "storage_account_name": {"key": "url", "parser": parse_storage_account_name},
}


def translate_abfs_spec(abfs_spec: dict | None) -> dict | None:
    """Translates an Azure Blob File Storage linked service in Data Factory's object model to a set of parameters
    used to connect to the ABFS location from a Databricks workspace.
    :parameter abfs_spec: Azure SQL Server linked service definition as a ``dict``
    :return: ABFS specification as a ``dict``
    """
    if abfs_spec is None:
        return None
    # Get the cluster properties:
    properties = abfs_spec.get("properties")
    # Translate the properties:
    translated_properties = translate(properties, mapping)
    result = {"service_name": abfs_spec.get("name")}
    if translated_properties is not None:
        result.update(translated_properties)
    return result



"""This module defines methods for translating Databricks Spark jar activities."""

from wkmigrate.utils import translate
from wkmigrate.activity_translators.parsers import (
    parse_dataset,
    parse_dataset_mapping,
    parse_dataset_properties,
)


mapping = {
    "source_properties": {"key": "source", "parser": parse_dataset_properties},
    "sink_properties": {"key": "sink", "parser": parse_dataset_properties},
    "column_mapping": {"key": "translator", "parser": parse_dataset_mapping},
    "source_dataset": {"key": "input_dataset_definitions", "parser": parse_dataset},
    "sink_dataset": {"key": "output_dataset_definitions", "parser": parse_dataset},
}


def translate_copy_activity(activity: dict) -> dict:
    """Translates a Databricks Spark jar activity definition in Data Factory's object model to a Databricks Spark jar
    task in the Databricks SDK object model.
    :parameter activity: Databricks Spark jar activity definition as a ``dict``
    :return: Databricks Spark jar task properties as a ``dict``
    """
    translated = translate(activity, mapping)
    if translated is None:
        raise ValueError('Translation failed')
    return translated



from wkmigrate.definition_stores.factory_definition_store import FactoryDefinitionStore
from wkmigrate.definition_stores.json_definition_store import JSONDefinitionStore
from wkmigrate.definition_stores.workspace_definition_store import (
    WorkspaceDefinitionStore,
)
from wkmigrate.definition_stores.yaml_definition_store import YAMLDefinitionStore

types = {
    "factory_definition_store": FactoryDefinitionStore,
    "json_definition_store": JSONDefinitionStore,
    "workspace_definition_store": WorkspaceDefinitionStore,
    "yaml_definition_store": YAMLDefinitionStore,
}



"""This module defines methods for translating If Condition activities."""

import warnings
from wkmigrate.activity_translators.parsers import parse_condition_expression
from wkmigrate.utils import translate

mapping = {
    "op": {
        "key": "expression",
        "parser": lambda x: parse_condition_expression(x).get("op"),
    },
    "left": {
        "key": "expression",
        "parser": lambda x: parse_condition_expression(x).get("left"),
    },
    "right": {
        "key": "expression",
        "parser": lambda x: parse_condition_expression(x).get("right"),
    },
}


def translate_if_condition_activity(activity: dict) -> dict | tuple[dict, list[dict]]:
    """Translates an If Condition activity definition in Data Factory's object model to a Databricks if/else condition
    task in the Databricks SDK object model.
    :parameter activity: If Condition activity definition as a ``dict``
    :return: Databricks if/else condition task properties as a ``dict``
    """
    # Parse the condition expression:
    translated_activity = translate(activity, mapping)
    if translated_activity is None:
        raise ValueError('Translation failed')
    if "if_false_activities" not in activity and "if_true_activities" not in activity:
        warnings.warn("No child activities of if-else condition activity", stacklevel=2)
        return translated_activity
    child_activities = []
    if "if_false_activities" in activity:
        if_false_activities = activity.get("if_false_activities")
        parent_task_name = activity.get("name")
        if if_false_activities is not None and parent_task_name is not None:
            child_activities.extend(
                parse_child_activities(
                    child_activities=if_false_activities,
                    parent_task_name=parent_task_name,
                    parent_task_outcome="false",
                )
            )
    if "if_true_activities" in activity:
        if_true_activities = activity.get("if_true_activities")
        parent_task_name = activity.get("name")
        if if_true_activities is not None and parent_task_name is not None:
            child_activities.extend(
                parse_child_activities(
                    child_activities=if_true_activities,
                    parent_task_name=parent_task_name,
                    parent_task_outcome="true",
                )
            )
    return translated_activity, child_activities


def parse_child_activities(child_activities: list[dict], parent_task_name: str, parent_task_outcome: str) -> list[dict]:
    """Translates if-else condition dependencies from Data Factory's object model to
    tasks in the Databricks SDK's object model.
    :parameter child_activities: List of child tasks as ``dict``
    :parameter parent_task_name: Name of the parent if-else condition task as a ``str``
    :parameter parent_task_outcome: Outcome of the parent if-else condition (either `"true"` or `"false`")
    :return: List of child tasks as ``dict``
    """
    translated_dependencies = []
    for activity in child_activities:
        if activity["depends_on"] is None or len(activity["depends_on"]) == 0:
            activity["depends_on"].append({"activity": parent_task_name, "outcome": parent_task_outcome})
        translated_dependencies.append(activity)
    return translated_dependencies



"""This module defines methods for translating Databricks schedule triggers from data pipelines."""

from wkmigrate.trigger_translators.parsers import parse_cron_expression
from wkmigrate.utils import translate


mapping = {
    "quartz_cron_expression": {"key": "recurrence", "parser": parse_cron_expression},
    "timezone_id": {"key": "time_zone", "parser": lambda x: "UTC"},
}


def translate_schedule_trigger(trigger_definition: dict) -> dict:
    """Translates a schedule trigger definition in Data Factory's object model to a Databricks cron
    schedule definition in the Databricks SDK object model.
    :parameter trigger_definition: Schedule trigger definition as a ``dict``
    :return: Databricks cron schedule definition as a ``dict``
    """
    # Get the properties:
    if "properties" not in trigger_definition:
        raise ValueError('No value for "properties" with trigger')
    properties = trigger_definition.get("properties")
    # Get the recurrence:
    if properties is None:
        raise ValueError('Properties cannot be None')
    if "recurrence" not in properties:
        raise ValueError('No value for "recurrence" with schedule trigger')
    # Translate the properties:
    translated = translate(properties, mapping)
    if translated is None:
        raise ValueError('Translation failed')
    return translated



"""This module defines methods for building ``DefinitionStore`` objects."""

from wkmigrate.definition_stores.definition_store import DefinitionStore
from wkmigrate.definition_stores import types


getters = types


def build_definition_store(definition_store_type: str, options: dict | None = None) -> DefinitionStore:
    """Gets a ``DefinitionStore`` object with the given options.
    :parameter definition_store_type: Definition store type
    :parameter options: A set of options for the specified definition store type
    :return: ``DefinitionStore``: A ``DefinitionStore`` of the specified type
    """
    getter = getters.get(definition_store_type, None)
    if getter is None:
        raise ValueError(f"No definition store registered with type {definition_store_type}")
    if options is None:
        raise ValueError(f"Options must be provided for definition store type {definition_store_type}")
    return getter(**options)



"""This module defines methods for parsing pipeline objects to the Databricks SDK's object model."""

from typing import Any


def parse_parameter_value(parameter_value: Any) -> Any:
    """Parses a value from a JSON string.
    :parameter parameter_value: Parameter value as a ``str``
    :return: Parsed parameter value with inferred data type
    """
    return str(parameter_value).replace("'", '"')



"""This module defines the ``JSONDefinitionStore`` class."""

import json
import os
from dataclasses import dataclass, field
from typing import Any
from collections.abc import Callable
from wkmigrate.definition_stores.definition_store import DefinitionStore
from wkmigrate.enums.json_source_type import JSONSourceType
from wkmigrate.pipeline_translators.pipeline_translator import translate_pipeline


@dataclass
class JSONDefinitionStore(DefinitionStore):
    """This class is used to create a target JSON file."""

    json_file_path: str
    json_source_type: JSONSourceType | None = JSONSourceType.DATA_FACTORY_PIPELINE
    _pipeline_getter: Callable[[list[dict], str], dict] | None = field(init=False)

    def __post_init__(self) -> None:
        """Validates the provided file path and sets the correct pipeline getter."""
        if not os.path.isfile(self.json_file_path):
            raise ValueError(f"No file found at path {self.json_file_path}")
        if self.json_source_type == JSONSourceType.DATABRICKS_WORKFLOW:
            self._pipeline_getter = JSONDefinitionStore._databricks_workspace_pipeline_getter
            return
        self._pipeline_getter = JSONDefinitionStore._factory_pipeline_getter

    def load(self, pipeline_name: str) -> dict:
        """Gets a dictionary representation of a Databricks workflow from the JSON file.
        :parameter pipeline_name: Name of the source pipeline as a ``str``
        :return: ``dict`` representation of the Databricks workflow
        """
        pipeline = self._get_pipeline(pipeline_name)
        if self.json_source_type == JSONSourceType.DATA_FACTORY_PIPELINE:
            return translate_pipeline(pipeline)
        return pipeline

    def dump(self, job_definition: dict) -> None:
        """Writes a workflow definition to the JSON file following the Databricks SDK ``Job`` object definition.
        :parameter job_definition: Databricks workflow definition as a ``dict``
        :return: ``None``
        """
        self._write_file(job_definition)

    def _read_file(self) -> Any:
        """Reads the JSONDefinitionStore file into an object.
        :return: Parsed JSON object.
        """
        with open(self.json_file_path, "rb") as json_file:
            return json.load(json_file)

    def _write_file(self, data: dict) -> None:
        """Writes an object to the JSONDefinitionStore file.
        :parameter data: Object to write to JSON as a ``dict``
        """
        with open(self.json_file_path, "wb") as json_file:
            return json.dump(data, json_file, sort_keys=False)

    def _get_pipeline(self, pipeline_name: str) -> dict:
        """Gets a pipeline definition with the provided pipeline name from the JSONDefinitionStore file.
        :parameter pipeline_name: Pipeline name as a ``str``
        :return: Pipeline definition as a ``dict``
        """
        source = self._read_file()
        if not isinstance(source, list):
            return source
        if self._pipeline_getter is None:
            raise ValueError("Pipeline getter is not initialized")
        return self._pipeline_getter(source, pipeline_name)

    @staticmethod
    def _factory_pipeline_getter(pipelines: list[dict], pipeline_name: str) -> dict:
        """Gets a pipeline definition from a list of definitions in Data Factory's object model.
        :parameter pipelines: Pipeline definitions as a ``list[dict]``
        :parameter pipeline_name: Pipeline name as a ``str``
        :return: Pipeline definition as a ``dict``
        """
        for pipeline in pipelines:
            if "name" not in pipeline:
                continue
            if pipeline.get("name") == pipeline_name:
                return pipeline
        raise ValueError(f'No pipeline found with name "{pipeline_name}"')

    @staticmethod
    def _databricks_workspace_pipeline_getter(pipelines: list[dict], pipeline_name: str) -> dict:
        """Gets a pipeline definition from a list of definitions in the Databricks SDK object model.
        :parameter pipelines: Pipeline definitions as a ``list[dict]``
        :parameter pipeline_name: Pipeline name as a ``str``
        :return: Pipeline definition as a ``dict``
        """
        for pipeline in pipelines:
            if "settings" not in pipeline:
                continue
            settings = pipeline.get("settings")
            if settings is None or "name" not in settings:
                continue
            if settings.get("name") == pipeline_name:
                return pipeline
        raise ValueError(f'No pipeline found with name "{pipeline_name}"')



"""Defines the ``FactoryClient`` and ``TestFactoryClient`` classes."""

import json
from abc import ABC, abstractmethod
from dataclasses import dataclass, field

from azure.identity import ClientSecretCredential
from azure.mgmt.datafactory import DataFactoryManagementClient

import wkmigrate


class FactoryClient(ABC):
    """A client implementing methods for getting data pipeline, linked service,
    dataset, and pipeline trigger definitions.
    """

    __test__ = False

    @abstractmethod
    def get_pipeline(self, pipeline_name: str) -> dict:
        pass

    @abstractmethod
    def get_trigger(self, pipeline_name: str) -> dict:
        pass

    @abstractmethod
    def get_dataset(self, dataset_name: str) -> dict:
        pass

    @abstractmethod
    def get_linked_service(self, linked_service_name: str) -> dict:
        pass


@dataclass
class FactoryManagementClient(FactoryClient):
    """A Data Factory management client implementing methods for getting data pipeline,
    linked service, dataset, and pipeline trigger definitions.
    """

    __test__ = False

    tenant_id: str
    client_id: str
    client_secret: str
    subscription_id: str
    resource_group_name: str
    factory_name: str
    management_client: DataFactoryManagementClient | None = field(init=False)

    def __post_init__(self) -> None:
        """Sets up the Data Factory management client for the provided credentials."""
        if self.tenant_id is None:
            raise ValueError("A tenant_id must be provided when creating a FactoryDefinitionStore")
        if self.client_id is None:
            raise ValueError("A client_id must be provided when creating a FactoryDefinitionStore")
        if self.client_secret is None:
            raise ValueError("A client_secret must be provided when creating a FactoryDefinitionStore")
        credential = ClientSecretCredential(
            tenant_id=self.tenant_id,
            client_id=self.client_id,
            client_secret=self.client_secret,
        )
        self.management_client = DataFactoryManagementClient(credential, self.subscription_id)

    def get_pipeline(self, pipeline_name: str) -> dict:
        """Gets a pipeline definition with the specified name.
        :parameter pipeline_name: Name of the Data Factory pipeline as a ``str``
        :return: Data Factory pipeline definition as a ``dict``
        """
        # Get the pipelines:
        if self.management_client is None:
            raise ValueError("management_client is not initialized")
        pipeline = self.management_client.pipelines.get(self.resource_group_name, self.factory_name, pipeline_name)
        # If no pipeline was found:
        if pipeline is None:
            raise ValueError(f'No pipeline found with name "{pipeline_name}"')
        return dict(pipeline.as_dict())

    def get_linked_service(self, linked_service_name: str) -> dict:
        """Gets a linked service with the specified name from a Data Factory.
        :parameter linked_service_name: Name of the linked service in Data Factory as a ``str``
        :return: Linked service definition as a ``dict``
        """
        # Get the linked service:
        if self.management_client is None:
            raise ValueError("management_client is not initialized")
        linked_service = self.management_client.linked_services.get(
            resource_group_name=self.resource_group_name,
            factory_name=self.factory_name,
            linked_service_name=linked_service_name,
        )
        # If no linked service was found:
        if linked_service is None:
            raise ValueError(f'No linked service found with name "{linked_service_name}"')
        return dict(linked_service.as_dict())

    def get_trigger(self, pipeline_name: str) -> dict:
        """Gets a single trigger for a Data Factory pipeline.
        :parameter pipeline_name: Name of the Data Factory pipeline as a ``str``
        :return: Triggers in the source Data Factory as a ``list[dict]``
        """
        # List the triggers:
        triggers = self._list_triggers()
        for trigger in triggers:
            # Get the trigger properties:
            properties = trigger.get("properties")
            if properties is None:
                continue
            # Get the associated pipeline definitions:
            pipelines = properties.get("pipelines")
            if pipelines is None:
                continue
            # Get the pipeline references:
            pipeline_references = [
                pipeline.get("pipeline_reference")
                for pipeline in pipelines
                if pipeline.get("pipeline_reference") is not None
            ]
            # Get the pipeline names:
            pipeline_names = [
                pipeline_reference.get("reference_name")
                for pipeline_reference in pipeline_references
                if pipeline_reference.get("reference_name") is not None
                and pipeline_reference.get("type") == "PipelineReference"
            ]
            # Get the trigger by pipeline name:
            if pipeline_name in pipeline_names:
                return trigger
        # If no trigger was found:
        raise ValueError(f'No trigger found for pipeline with name "{pipeline_name}"')

    def _list_triggers(self) -> list[dict]:
        """Lists triggers in the source Data Factory.
        :return: Triggers in the source Data Factory as a ``list[dict]``
        """
        # List the triggers:
        if self.management_client is None:
            raise ValueError("management_client is not initialized")
        triggers = self.management_client.triggers.list_by_factory(
            resource_group_name=self.resource_group_name, factory_name=self.factory_name
        )
        # If no triggers were found:
        if triggers is None:
            raise ValueError(f'No triggers found for factory "{self.factory_name}"')
        return [dict(trigger.as_dict()) for trigger in triggers]

    def get_dataset(self, dataset_name: str) -> dict:
        """Gets a single dataset from a source Data Factory.
        :parameter dataset_name: Dataset name as a ``str``
        :return: Dataset definition as a ``dict``
        """
        # List the datasets:
        datasets = self._list_datasets()
        for dataset in datasets:
            # Check the dataset name:
            if dataset.get("name") != dataset_name:
                continue
            # Get the dataset properties:
            properties = dataset.get("properties")
            if properties is None:
                return dataset
            # Get the associated linked service:
            linked_service = properties.get("linked_service_name")
            if linked_service is None:
                return dataset
            # Get the linked service reference name:
            linked_service_name = linked_service.get("reference_name")
            # Get the linked service definition:
            linked_service_definition = self.get_linked_service(linked_service_name)
            # Append the linked service definition to the dataset object:
            dataset["linked_service_definition"] = linked_service_definition
            return dataset
        # If no datasets were found:
        raise ValueError(f'No dataset found for factory with name "{dataset_name}"')

    def _list_datasets(self) -> list[dict]:
        """Lists datasets in the source Data Factory.
        :return: Datasets in the source Data Factory as a ``list[dict]``
        """
        # List the datasets:
        if self.management_client is None:
            raise ValueError("management_client is not initialized")
        datasets = self.management_client.datasets.list_by_factory(
            resource_group_name=self.resource_group_name, factory_name=self.factory_name
        )
        # If no datasets were found:
        if datasets is None:
            raise ValueError(f'No datasets found for factory "{self.factory_name}"')
        return [dict(dataset.as_dict()) for dataset in datasets]


@dataclass
class FactoryTestClient(FactoryClient):
    """A mock client implementing methods for getting data pipeline, linked service,
    dataset, and pipeline trigger definitions.
    """

    test_json_path: str = wkmigrate.JSON_PATH

    def get_pipeline(self, pipeline_name: str) -> dict:
        """Gets a pipeline definition with the specified name.
        :parameter pipeline_name: Name of the Data Factory pipeline as a ``str``
        :return: Data Factory pipeline definition as a ``dict``
        """
        # Open the test pipelines file:
        with open(f"{self.test_json_path}/test_pipelines.json", "rb") as file:
            # Load the data from JSON:
            pipelines = json.load(file)
        # Get the pipeline by name:
        for pipeline in pipelines:
            if pipeline.get("name") == pipeline_name:
                return pipeline
        # If no pipeline was found:
        raise ValueError(f'No pipeline found with name "{pipeline_name}"')

    def get_trigger(self, pipeline_name: str) -> dict:
        """Gets a single trigger for a Data Factory pipeline.
        :parameter pipeline_name: Name of the Data Factory pipeline as a ``str``
        :return: Triggers in the source Data Factory as a ``list[dict]``
        """
        # Open the test triggers file:
        with open(f"{self.test_json_path}/test_triggers.json", "rb") as file:
            # Load the data from JSON:
            triggers = json.load(file)
        # Get the pipeline by name:
        for trigger in triggers:
            # Get the trigger properties:
            properties = trigger.get("properties")
            if properties is None:
                continue
            # Get the associated pipeline definitions:
            pipelines = properties.get("pipelines")
            if pipelines is None:
                continue
            # Get the pipeline references:
            pipeline_references = [
                pipeline.get("pipeline_reference")
                for pipeline in pipelines
                if pipeline.get("pipeline_reference") is not None
            ]
            # Get the pipeline names:
            pipeline_names = [
                pipeline_reference.get("reference_name")
                for pipeline_reference in pipeline_references
                if (
                    pipeline_reference.get("reference_name") is not None
                    and pipeline_reference.get("type") == "PipelineReference"
                )
            ]
            # Get the trigger by pipeline name:
            if pipeline_name in pipeline_names:
                return trigger
        # If no trigger was found:
        raise ValueError(f'No trigger found for pipeline with name "{pipeline_name}"')

    def get_dataset(self, dataset_name: str) -> dict:
        """Gets a single dataset from a source Data Factory.
        :parameter dataset_name: Dataset name as a ``str``
        :return: Dataset definition as a ``dict``
        """
        # Open the test datasets file:
        with open(f"{self.test_json_path}/test_datasets.json", "rb") as file:
            # Load the data from JSON:
            datasets = json.load(file)
        for dataset in datasets:
            # Get the dataset properties:
            properties = dataset.get("properties")
            if properties is None:
                return dataset
            # Get the associated linked service:
            linked_service = properties.get("linked_service_name")
            if linked_service is None:
                return dataset
            # Get the linked service reference name:
            linked_service_name = linked_service.get("reference_name")
            # Get the linked service definition:
            linked_service_definition = self.get_linked_service(linked_service_name)
            # Append the linked service definition to the dataset object:
            dataset["linked_service_definition"] = linked_service_definition
            return dataset
        # If no datasets were found:
        raise ValueError(f'No dataset found for factory with name "{dataset_name}"')

    def get_linked_service(self, linked_service_name: str) -> dict:
        """Gets a linked service with the specified name from a Data Factory.
        :parameter linked_service_name: Name of the linked service in Data Factory as a ``str``
        :return: Linked service definition as a ``dict``
        """
        # Open the test linked_services file:
        with open(f"{self.test_json_path}/test_linked_services.json", "rb") as file:
            # Load the data from JSON:
            linked_services = json.load(file)
        # Get the linked_service by name:
        for linked_service in linked_services:
            if linked_service.get("name") == linked_service_name:
                return linked_service
        # If no linked service was found:
        raise ValueError(f'No linked service found with name "{linked_service_name}"')



from enum import StrEnum


class IsolationLevel(StrEnum):
    ReadCommitted = ("READ_COMMITTED",)
    ReadUncommitted = ("READ_UNCOMMITTED",)
    RepeatableRead = ("REPEATABLE_READ",)
    Serializable = ("SERIALIZABLE",)
    Snapshot = "SNAPSHOT"



from wkmigrate.datasets.parsers import (
    parse_avro_file_properties,
    parse_avro_file_dataset,
    parse_delimited_file_dataset,
    parse_delimited_file_properties,
    parse_delta_properties,
    parse_delta_table_dataset,
    parse_json_file_dataset,
    parse_json_file_properties,
    parse_orc_file_properties,
    parse_orc_file_dataset,
    parse_parquet_file_properties,
    parse_parquet_file_dataset,
    parse_sql_server_properties,
    parse_sql_server_dataset,
)


secrets = {
    "avro": ["storage_account_key"],
    "csv": ["storage_account_key"],
    "delta": [],
    "json": ["storage_account_key"],
    "orc": ["storage_account_key"],
    "parquet": ["storage_account_key"],
    "sqlserver": ["host", "database", "user_name", "password"],
}


options = {
    "csv": [
        "header",
        "sep",
        "lineSep",
        "quote",
        "quoteAll",
        "escape",
        "nullValue",
        "compression",
        "encoding",
    ],
    "json": ["encoding", "compression"],
    "orc": ["compression"],
    "parquet": ["compression"],
    "sqlserver": ["host", "database", "user_name", "password"],
}


dataset_parsers = {
    "Avro": parse_avro_file_dataset,
    "AzureDatabricksDeltaLakeDataset": parse_delta_table_dataset,
    "AzureSqlTable": parse_sql_server_dataset,
    "DelimitedText": parse_delimited_file_dataset,
    "Json": parse_json_file_dataset,
    "Orc": parse_orc_file_dataset,
    "Parquet": parse_parquet_file_dataset,
}


property_parsers = {
    "AvroSource": parse_avro_file_properties,
    "AvroSink": parse_avro_file_properties,
    "AzureDatabricksDeltaLakeSource": parse_delta_properties,
    "AzureDatabricksDeltaLakeSink": parse_delta_properties,
    "AzureSqlSource": parse_sql_server_properties,
    "AzureSqlSink": parse_sql_server_properties,
    "DelimitedTextSource": parse_delimited_file_properties,
    "DelimitedTextSink": parse_delimited_file_properties,
    "JsonSource": parse_json_file_properties,
    "JsonSink": parse_json_file_properties,
    "OrcSource": parse_orc_file_properties,
    "OrcSink": parse_orc_file_properties,
    "ParquetSource": parse_parquet_file_properties,
    "ParquetSink": parse_parquet_file_properties,
}



from wkmigrate.activity_translators.copy_activity_translator import (
    translate_copy_activity,
)
from wkmigrate.activity_translators.for_each_activity_translator import (
    translate_for_each_activity,
)
from wkmigrate.activity_translators.if_condition_activity_translator import (
    translate_if_condition_activity,
)
from wkmigrate.activity_translators.notebook_activity_translator import (
    translate_notebook_activity,
)
from wkmigrate.activity_translators.spark_jar_activity_translator import (
    translate_spark_jar_activity,
)
from wkmigrate.activity_translators.spark_python_activity_translator import (
    translate_spark_python_activity,
)


type_mapping = {
    "DatabricksNotebook": "notebook_task",
    "DatabricksSparkJar": "spark_jar_task",
    "DatabricksSparkPython": "spark_python_task",
    "IfCondition": "condition_task",
    "ForEach": "for_each_task",
    "Copy": "copy_data_task",
}

type_translators = {
    "DatabricksNotebook": translate_notebook_activity,
    "DatabricksSparkJar": translate_spark_jar_activity,
    "DatabricksSparkPython": translate_spark_python_activity,
    "IfCondition": translate_if_condition_activity,
    "ForEach": translate_for_each_activity,
    "Copy": translate_copy_activity,
}



"""Defines ``WorkspaceClient`` classes."""

import base64
import json
import warnings
from abc import ABC, abstractmethod
from dataclasses import dataclass, field

import autopep8  # type: ignore
from databricks.sdk import WorkspaceClient
from databricks.sdk.service.jobs import BaseJob, CronSchedule, Job, Task
from databricks.sdk.service.pipelines import NotebookLibrary, PipelineLibrary
from databricks.sdk.service.workspace import ExportFormat, ImportFormat, Language

import wkmigrate
from wkmigrate.datasets import options, secrets


class DatabricksWorkspaceClient(ABC):
    """A client implementing methods for getting data pipeline, linked service,
    dataset, and pipeline trigger definitions.
    """

    __test__ = False

    @abstractmethod
    def get_workflow(self, job_id: int | None = None, job_name: str | None = None) -> Job:
        pass

    @abstractmethod
    def create_workflow(self, job_definition: dict, translation_options: dict | None = None) -> int | None:
        pass


@dataclass
class WorkspaceManagementClient(DatabricksWorkspaceClient):
    """A client used to list, describe, and update objects in a Databricks workspace."""

    __test__ = False

    authentication_type: str
    host_name: str
    pat: str | None = None
    username: str | None = None
    password: str | None = None
    resource_id: str | None = None
    tenant_id: str | None = None
    client_id: str | None = None
    client_secret: str | None = None
    workspace_client: WorkspaceClient | None = field(init=False)

    def __post_init__(self) -> None:
        """Sets up the workspace client for the provided authentication credentials."""
        if self.authentication_type == "pat":
            self.workspace_client = self._login_with_pat()
            return
        if self.authentication_type == "basic":
            self.workspace_client = self._login_with_basic_auth()
            return
        if self.authentication_type == "azure-client-secret":
            self.workspace_client = self._login_with_azure_client_secret()
            return
        raise ValueError(
            'Got an invalid value for "self.authentication_type", must be "pat", "basic", or "azure-client-secret"'
        )

    def _login_with_pat(self) -> WorkspaceClient:
        """Creates a ``WorkspaceClient`` with PAT authentication.
        :return: A ``WorkspaceClient`` from the Databricks SDK
        """
        if self.pat is None:
            raise ValueError('No value provided for "pat" with access token authentication')
        return WorkspaceClient(auth_type=self.authentication_type, host=self.host_name, token=self.pat)

    def _login_with_basic_auth(self) -> WorkspaceClient:
        """Creates a ``WorkspaceClient`` with basic authentication.
        :return: A ``WorkspaceClient`` from the Databricks SDK
        """
        if self.username is None:
            raise ValueError('No value provided for "username" with basic authentication')
        if self.password is None:
            raise ValueError('No value provided for "password" with basic authentication')
        return WorkspaceClient(
            auth_type=self.authentication_type,
            host=self.host_name,
            username=self.username,
            password=self.password,
        )

    def _login_with_azure_client_secret(self) -> WorkspaceClient:
        """Creates a ``WorkspaceClient`` with Azure client secret authentication.
        :return: A ``WorkspaceClient`` from the Databricks SDK
        """
        if self.resource_id is None:
            raise ValueError('No value provided for "resource_id" with Azure client secret authentication')
        if self.tenant_id is None:
            raise ValueError('No value provided for "tenant_id" with Azure client secret authentication')
        if self.client_id is None:
            raise ValueError('No value provided for "client_id" with Azure client secret authentication')
        if self.client_secret is None:
            raise ValueError('No value provided for "client_secret" with Azure client secret authentication')
        return WorkspaceClient(
            auth_type=self.authentication_type,
            host=self.host_name,
            azure_workspace_resource_id=self.resource_id,
            azure_tenant_id=self.tenant_id,
            client_id=self.client_id,
            client_secret=self.client_secret,
        )

    def get_workflow(self, job_id: int | None = None, job_name: str | None = None) -> Job:
        """Gets a workflow with the specified ID or name as a ``Job`` object.
        :parameter job_id: Job ID for the specified workflow
        :parameter job_name: Job name for the specified workflow
        :return: Workflow definition as a ``Job``
        """
        # Check the input parameters:
        if job_id is None and job_name is None:
            raise ValueError('Must provide a value for "job_id" or "job_name".')
        # If a workflow ID is specified, get the workflow by ID:
        if job_id is not None:
            if self.workspace_client is None:
                raise ValueError("workspace_client is not initialized")
            return self.workspace_client.jobs.get(job_id=job_id)
        # Otherwise, list the workflows by name:
        if job_name is None:
            raise ValueError("job_name cannot be None when job_id is not provided")
        workflows = self._list_workflows_by_name(job_name=job_name)
        # If more than 1 workflow exists with the specified name:
        if len(workflows) > 1:
            raise ValueError(f'Duplicate workflows found in the target workspace with name "{job_name}"')
        if self.workspace_client is None:
            raise ValueError("workspace_client is not initialized")
        job_id = workflows[0].job_id
        if job_id is None:
            raise ValueError("Job ID cannot be None")
        return self.workspace_client.jobs.get(job_id=job_id)

    def _list_workflows_by_name(self, job_name: str) -> list[BaseJob]:
        """Gets workflows with the specified name as ``BaseJob`` objects.
        :return: Workflow definitions as a ``list[BaseJob]``
        """
        # List the workflows:
        if self.workspace_client is None:
            raise ValueError("workspace_client is not initialized")
        workflows = list(self.workspace_client.jobs.list(name=job_name))
        # If no workflows were found:
        if workflows is None or len(workflows) == 0:
            raise ValueError(f'No workflows found in the target workspace with name "{job_name}"')
        return workflows

    def create_workflow(self, job_definition: dict, translation_options: dict | None = None) -> int:
        """Creates a workflow with the specified definition and options.
        :parameter job_definition: Workflow definition settings as a ``dict``
        :parameter translation_options: Workflow translation options as a ``dict``
        :return: Created Job ID as an ``int``
        """
        job_settings = job_definition.get("settings")
        if job_settings is None:
            raise ValueError('Invalid "job_definition" object.')
        job_name = job_settings.get("name", None)
        if job_name is None:
            raise ValueError('No value provided for "name"')

        # access_control_list = job_settings.get("access_control_list", None)
        # is_continuous = job_settings.get("is_continuous", None)
        # deployment = job_settings.get("deployment", None)
        # edit_mode = job_settings.get("edit_mode", None)
        # email_notifications = job_settings.get("email_notifications", None)
        # git_source = job_settings.get("git_source", None)
        # health_rules = job_settings.get("health_rules", None)
        # max_concurrent_runs = job_settings.get("max_concurrent_runs", None)
        # notification_settings = job_settings.get("notification_settings", None)
        # parameter_definitions = job_settings.get("parameter_definitions", None)
        # queue = job_settings.get("queue", None)
        # run_as_principal = job_settings.get("run_as_principal", None)
        # webhook_notifications = job_settings.get("webhook_notifications", None)
        if self.workspace_client is None:
            raise ValueError("workspace_client is not initialized")
        tasks = job_settings.get("tasks")
        if tasks is None:
            raise ValueError('No "tasks" provided in job_settings')
        response = self.workspace_client.jobs.create(
            name=job_settings.get("name"),
            description=job_settings.get("description"),
            schedule=WorkspaceManagementClient._get_schedule(job_settings.get("schedule")),
            tags=job_settings.get("tags"),
            tasks=[self._create_task(task, translation_options) for task in tasks],
            timeout_seconds=job_settings.get("timeout_seconds"),
        )
        job_id = response.job_id
        if job_id is None:
            raise ValueError("Created job ID cannot be None")
        return job_id

    def _create_task(self, task: dict, translation_options: dict | None) -> Task:
        """Creates a Databricks workflow ``Task`` object from the task definition.
        :parameter task: Workflow task definition as a ``dict``
        :parameter translation_options: Workflow translation options as a ``dict``
        :return: Workflow ``Task`` object
        """
        if "type" not in task:
            raise ValueError('Task has no "type"')
        if task.get("type") == "DatabricksNotebook":
            self._create_notebook_task_dependencies(task)
        if task.get("type") == "Copy":
            files_to_delta_sinks = True
            if translation_options is not None:
                files_to_delta_sinks = translation_options.get("files_to_delta_sinks", True)
            dependency_object = self._create_copy_task_dependencies(task, files_to_delta_sinks)
            task.pop("copy_data_task")
            if not files_to_delta_sinks:
                task["notebook_task"] = {"notebook_path": dependency_object}
            else:
                task["pipeline_task"] = {"pipeline_id": dependency_object}
        if task.get("type") == "ForEach":
            return self._create_for_each_task(task)
        return Task.from_dict(task)

    def _create_copy_task_dependencies(self, task: dict, files_to_delta_sinks: bool) -> str:
        """Creates a Databricks notebook to copy data.
        :parameter task: Workflow task definition as a ``dict``
        :parameter files_to_delta_sinks: Whether to create Lakeflow Declarative Pipelines to copy data files to Delta table sinks
        :return: Databricks notebook path as a ``str``
        """
        copy_data_task = task.get("copy_data_task")
        if copy_data_task is None:
            raise ValueError('No "copy_data_task" found in task')
        source_dataset = copy_data_task.get("source_dataset")
        source_properties = copy_data_task.get("source_properties")
        if source_dataset is None or source_properties is None:
            raise ValueError('Missing "source_dataset" or "source_properties"')
        source_definition = {
            **source_dataset,
            **source_properties,
        }
        self._create_data_source_secrets(source_definition)
        sink_dataset = copy_data_task.get("sink_dataset")
        sink_properties = copy_data_task.get("sink_properties")
        if sink_dataset is None or sink_properties is None:
            raise ValueError('Missing "sink_dataset" or "sink_properties"')
        sink_definition = {
            **sink_dataset,
            **sink_properties,
        }
        self._create_data_source_secrets(sink_definition)
        column_mapping = copy_data_task.get("column_mapping")
        notebook_path = self._create_copy_data_notebook(
            source_definition, sink_definition, column_mapping, files_to_delta_sinks
        )
        if not files_to_delta_sinks:
            return notebook_path
        # Create a DLT pipeline from the notebook path:
        if self.workspace_client is None:
            raise ValueError("workspace_client is not initialized")
        pipeline_response = self.workspace_client.pipelines.create(
            allow_duplicate_names=True,
            catalog="wkmigrate",
            channel="CURRENT",
            continuous=False,
            development=False,
            libraries=[PipelineLibrary(notebook=NotebookLibrary(path=notebook_path))],
            name=f'{task.get("task_key")}_pipeline',
            photon=True,
            serverless=True,
            target="wkmigrate",
        )
        pipeline_id = pipeline_response.pipeline_id
        if pipeline_id is None:
            raise ValueError("Created pipeline ID cannot be None")
        return pipeline_id

    def _create_copy_data_notebook(
        self,
        source_definition: dict,
        sink_definition: dict,
        column_mapping: dict,
        files_to_delta_sinks: bool,
    ) -> str:
        """Creates a notebook in the target workspace to copy data between specified data source and sink with
        the given column mapping.
        :parameter source_definition: Source dataset definition as a ``dict``
        :parameter sink_definition: Sink dataset definition as a ``dict``
        :parameter column_mapping: Column-level mapping as a ``dict``
        :parameter files_to_delta_sinks: A ``bool`` indicating whether file sinks should be
                                            converted to Delta tables
        :return: Databricks notebook path in the target workspace as a ``str``
        """
        script_lines = [
            "# Databricks notebook source",
            "import pyspark.sql.types as T",
            "import pyspark.sql.functions as F",
            "",
            "# Set the source options:",
        ]
        # Append code blocks to set source dataset options from the Databricks secret scope:
        script_lines.extend(WorkspaceManagementClient._get_option_expressions(source_definition))
        if not files_to_delta_sinks:
            # Append code blocks to set source dataset options from the Databricks secret scope:
            script_lines.append("# Set the target options:")
            script_lines.extend(WorkspaceManagementClient._get_option_expressions(sink_definition))
            # Append a code block to read the source as a DataFrame:
            script_lines.append("# Read from the source:")
            script_lines.append(WorkspaceManagementClient._get_read_expression(source_definition))
            # Append code blocks to create a new DataFrame with mapped column names and data types:
            script_lines.append("# Map the source columns to the target columns:")
            script_lines.append(
                WorkspaceManagementClient._get_mapping(source_definition, sink_definition, column_mapping)
            )
            # Append a code block to write the DataFrame to Delta:
            script_lines.append("# Write to the target:")
            script_lines.append(WorkspaceManagementClient._get_write_expression(sink_definition))
        else:
            # Append a code block to define the Delta table in DLT:
            script_lines.append("# Load the data with DLT as a materialized view:")
            script_lines.append(
                WorkspaceManagementClient._get_dlt_definition(source_definition, sink_definition, column_mapping)
            )
        # Create and upload the script as a Python notebook:
        source_dataset_name = source_definition.get("dataset_name")
        sink_dataset_name = sink_definition.get("dataset_name")
        notebook_str = autopep8.fix_code("\n".join(script_lines))
        notebook_path = f"/wkmigrate/copy_data_notebooks/copy_{source_dataset_name}_to_{sink_dataset_name}"
        if self.workspace_client is None:
            raise ValueError("workspace_client is not initialized")
        self.workspace_client.workspace.mkdirs("/wkmigrate/copy_data_notebooks")
        self.workspace_client.workspace.import_(
            content=base64.b64encode(notebook_str.encode()).decode(),
            format=ImportFormat.SOURCE,
            language=Language.PYTHON,
            overwrite=True,
            path=notebook_path,
        )
        return notebook_path

    @staticmethod
    def _get_dlt_definition(source_dataset: dict, sink_dataset: dict, column_mapping: dict) -> str:
        """Creates a DLT table definition for a give source, sink, and column mapping.
        :parameter source_dataset: Source dataset properties as a ``dict``
        :parameter sink_dataset: Sink dataset properties as a ``dict``
        :parameter column_mapping: Column mapping as a ``dict``
        :return: A DLT table definition in Python as a ``str``
        """
        source_name = source_dataset.get("dataset_name")
        sink_name = sink_dataset.get("dataset_name")
        return f"""@dlt.table(
                        name="{sink_name}",
                        comment="Data copied from {source_name}; Previously targeted {sink_name}."
                        tbl_properties={{'delta.createdBy.wkmigrate': 'true'}}
                    )
                    def {sink_name}:
                        {WorkspaceManagementClient._get_read_expression(source_dataset)}
                        {WorkspaceManagementClient._get_mapping(source_dataset, sink_dataset, column_mapping)}
                        return {sink_name}_df
                """

    @staticmethod
    def _get_mapping(source_dataset: dict, sink_dataset: dict, column_mapping: dict) -> str:
        """Creates a PySpark expression mapping columns from a source DataFrame to a sink DataFrame.
        :parameter source_dataset: Source dataset properties as a ``dict``
        :parameter sink_dataset: Sink dataset properties as a ``dict``
        :parameter column_mapping: Column mapping as a ``dict``
        :return: PySpark expression creating a new DataFrame with the mapped columns
        """
        source_name = source_dataset.get("dataset_name")
        sink_name = sink_dataset.get("dataset_name")
        mapping_expressions = [
            f'"cast({mapping["source_column_name"]} as {mapping["sink_column_type"]}) as {mapping["sink_column_name"]}"'
            for mapping in column_mapping
        ]
        newline_characters = ", \n\t"
        return f"{sink_name}_df = {source_name}_df.selectExpr(\n\t{newline_characters.join(mapping_expressions)}\n)"

    @staticmethod
    def _get_write_expression(sink_definition: dict) -> str:
        """Creates a PySpark expression writing to a specified data sink.
        :parameter sink_definition: Sink dataset properties as a ``dict``
        :return: PySpark expression writing a DataFrame to the sink
        """
        sink_name = sink_definition.get("dataset_name")
        sink_type = sink_definition.get("type")
        if sink_type == "avro":
            container_name = sink_definition.get("container")
            storage_account_name = sink_definition.get("storage_account_name")
            folder_path = sink_definition.get("folder_path")
            return rf"""{sink_name}_df.write.format("avro")  \
                        .mode("overwrite")  \
                        .save("abfss://{container_name}@{storage_account_name}.dfs.core.windows.net/{folder_path}")
                    """
        if sink_type == "csv":
            container_name = sink_definition.get("container")
            storage_account_name = sink_definition.get("storage_account_name")
            folder_path = sink_definition.get("folder_path")
            return rf"""{sink_name}_df.write.format("csv")  \
                        .options(**{sink_name}_options)  \
                        .mode("overwrite")  \
                        .save("abfss://{container_name}@{storage_account_name}.dfs.core.windows.net/{folder_path}")
                    """
        if sink_type == "delta":
            database_name = sink_definition.get("database_name")
            table_name = sink_definition.get("table_name")
            return rf"""{sink_name}_df.write.format("delta")  \
                        .mode("overwrite")  \
                        .saveAsTable("hive_metastore.{database_name}.{table_name}")
                    """
        if sink_type == "json":
            container_name = sink_definition.get("container")
            storage_account_name = sink_definition.get("storage_account_name")
            folder_path = sink_definition.get("folder_path")
            return rf"""{sink_name}_df.write.format("json")  \
                        .options(**{sink_name}_options)  \
                        .mode("overwrite")  \
                        .save("abfss://{container_name}@{storage_account_name}.dfs.core.windows.net/{folder_path}")
                    """
        if sink_type == "orc":
            container_name = sink_definition.get("container")
            storage_account_name = sink_definition.get("storage_account_name")
            folder_path = sink_definition.get("folder_path")
            return rf"""{sink_name}_df.write.format("orc")  \
                        .options(**{sink_name}_options)  \
                        .mode("overwrite")  \
                        .save("abfss://{container_name}@{storage_account_name}.dfs.core.windows.net/{folder_path}")
                    """
        if sink_type == "parquet":
            container_name = sink_definition.get("container")
            storage_account_name = sink_definition.get("storage_account_name")
            folder_path = sink_definition.get("folder_path")
            return rf"""{sink_name}_df.write.format("parquet")  \
                        .options(**{sink_name}_options)  \
                        .mode("overwrite")  \
                        .save("abfss://{container_name}@{storage_account_name}.dfs.core.windows.net/{folder_path}")
                    """
        if sink_type == "sqlserver":
            return ""  # TODO: SET UP A WRITE TO SQL SERVER
        raise ValueError(f'Writing data to "{sink_type}" not supported')

    @staticmethod
    def _get_read_expression(source_definition: dict) -> str:
        """Creates a PySpark expression reading from a specified data source.
        :parameter source_definition: Source dataset properties as a ``dict``
        :return: PySpark expression reading from the source to a DataFrame
        """
        source_name = source_definition.get("dataset_name")
        source_type = source_definition.get("type")
        if source_type == "avro":
            container_name = source_definition.get("container")
            storage_account_name = source_definition.get("storage_account_name")
            folder_path = source_definition.get("folder_path")
            return f"""{source_name}_df = ( 
                        spark.read.format("avro")
                            .load("abfss://{container_name}@{storage_account_name}.dfs.core.windows.net/{folder_path}")
                    )
                    """
        if source_type == "csv":
            container_name = source_definition.get("container")
            storage_account_name = source_definition.get("storage_account_name")
            folder_path = source_definition.get("folder_path")
            return f"""{source_name}_df = ( 
                        spark.read.format("csv")
                            .options(**{source_name}_options)
                            .load("abfss://{container_name}@{storage_account_name}.dfs.core.windows.net/{folder_path}")
                        )
                    """
        if source_type == "delta":
            database_name = source_definition.get("database_name")
            table_name = source_definition.get("table_name")
            return f'{source_name}_df = spark.read.table("hive_metastore.{database_name}.{table_name}'
        if source_type == "json":
            container_name = source_definition.get("container")
            storage_account_name = source_definition.get("storage_account_name")
            folder_path = source_definition.get("folder_path")
            return f"""{source_name}_df = ( 
                        spark.read.format("json")
                            .options(**{source_name}_options)
                            .load("abfss://{container_name}@{storage_account_name}.dfs.core.windows.net/{folder_path}")
                        )
                    """
        if source_type == "orc":
            container_name = source_definition.get("container")
            storage_account_name = source_definition.get("storage_account_name")
            folder_path = source_definition.get("folder_path")
            return f"""{source_name}_df = ( 
                        spark.read.format("orc")
                            .options(**{source_name}_options)
                            .load("abfss://{container_name}@{storage_account_name}.dfs.core.windows.net/{folder_path}")
                        )
                    """
        if source_type == "parquet":
            container_name = source_definition.get("container")
            storage_account_name = source_definition.get("storage_account_name")
            folder_path = source_definition.get("folder_path")
            return f"""{source_name}_df = ( 
                        spark.read.format("parquet")
                            .options(**{source_name}_options)
                            .load("abfss://{container_name}@{storage_account_name}.dfs.core.windows.net/{folder_path}")
                        )
                    """
        if source_type == "sqlserver":
            schema_name = source_definition.get("schema_name")
            table_name = source_definition.get("table_name")
            return f"""{source_name}_df = ( 
                    spark.read.format("sqlserver")
                        .options(**{source_name}_options)
                        .option("dbtable", "{schema_name}.{table_name}")
                        .load()
                    )
                    """
        raise ValueError(f'Reading data from "{source_type}" not supported')

    @staticmethod
    def _get_option_expressions(dataset_definition: dict) -> list[str]:
        """Creates a PySpark expression defining DataFrameReader or DataFrameWriter options.
        :parameter dataset_definition: Dataset properties as a ``dict``
        :return: PySpark expression writing a DataFrame to the sink
        """
        dataset_name = dataset_definition.get("dataset_name")
        service_name = dataset_definition.get("service_name")
        dataset_type = dataset_definition.get("type")
        if dataset_type == "avro":
            config_lines = [
                f"""spark.conf.set(
                    "fs.azure.account.key.{dataset_definition.get('storage_account_name')}.dfs.core.windows.net",
                    dbutils.secrets.get(
                        scope="wkmigrate_credentials_scope", 
                        key="{service_name}_storage_account_key"
                    )
                )""",
                f"""spark.conf.set(
                    "spark.sql.files.maxRecordsPerFile",
                    "{dataset_definition.get('records_per_file')}"
                )""",
            ]
            return [f"{dataset_name}_options = {{}}", *config_lines]
        if dataset_type == "csv":
            config_lines = [
                rf'{dataset_name}_options["{option}"] = r"{dataset_definition.get(option)}"'
                for option in options[dataset_type]
                if dataset_definition.get(option)
            ]
            if "records_per_file" in dataset_definition:
                records_per_file = dataset_definition.get("records_per_file")
                config_lines.append(f'spark.conf.set("spark.sql.files.maxRecordsPerFile", "{records_per_file}")')
            config_lines.append(
                f"""spark.conf.set(
                    "fs.azure.account.key.{dataset_definition.get('storage_account_name')}.dfs.core.windows.net",
                        dbutils.secrets.get(
                            scope="wkmigrate_credentials_scope", 
                            key="{service_name}_storage_account_key"
                    )
                )
                """
            )
            return [f"{dataset_name}_options = {{}}", *config_lines]
        if dataset_type == "json":
            config_lines = [
                rf'{dataset_name}_options["{option}"] = r"{dataset_definition.get(option)}"'
                for option in options[dataset_type]
                if dataset_definition.get(option)
            ]
            if "records_per_file" in dataset_definition:
                records_per_file = dataset_definition.get("records_per_file")
                config_lines.append(f'spark.conf.set("spark.sql.files.maxRecordsPerFile", "{records_per_file}")')
            config_lines.append(
                f"""spark.conf.set(
                    "fs.azure.account.key.{dataset_definition.get('storage_account_name')}.dfs.core.windows.net",
                        dbutils.secrets.get(
                            scope="wkmigrate_credentials_scope", 
                            key="{service_name}_storage_account_key"
                    )
                )
                """
            )
            return [f"{dataset_name}_options = {{}}", *config_lines]
        if dataset_type == "orc":
            config_lines = [
                rf'{dataset_name}_options["{option}"] = r"{dataset_definition.get(option)}"'
                for option in options[dataset_type]
                if dataset_definition.get(option)
            ]
            if "records_per_file" in dataset_definition:
                records_per_file = dataset_definition.get("records_per_file")
                config_lines.append(f'spark.conf.set("spark.sql.files.maxRecordsPerFile", "{records_per_file}")')
            config_lines.append(
                f"""spark.conf.set(
                    "fs.azure.account.key.{dataset_definition.get('storage_account_name')}.dfs.core.windows.net",
                        dbutils.secrets.get(
                            scope="wkmigrate_credentials_scope", 
                            key="{service_name}_storage_account_key"
                    )
                )
                """
            )
            return [f"{dataset_name}_options = {{}}", *config_lines]
        if dataset_type == "parquet":
            config_lines = [
                rf'{dataset_name}_options["{option}"] = r"{dataset_definition.get(option)}"'
                for option in options[dataset_type]
                if dataset_definition.get(option)
            ]
            if "records_per_file" in dataset_definition:
                records_per_file = dataset_definition.get("records_per_file")
                config_lines.append(f'spark.conf.set("spark.sql.files.maxRecordsPerFile", "{records_per_file}")')
            config_lines.append(
                f"""spark.conf.set(
                    "fs.azure.account.key.{dataset_definition.get('storage_account_name')}.dfs.core.windows.net",
                        dbutils.secrets.get(
                            scope="wkmigrate_credentials_scope", 
                            key="{service_name}_storage_account_key"
                    )
                )
                """
            )
            return [f"{dataset_name}_options = {{}}", *config_lines]
        if dataset_type == "sqlserver":
            config_lines = [
                f"""{dataset_name}_options["{option}"] = dbutils.secrets.get(
                    scope="wkmigrate_credentials_scope", 
                    key="{service_name}_{option}"
                )
                """
                for option in options[dataset_type]
            ]
            return [f"{dataset_name}_options = {{}}", *config_lines]
        return []

    def _create_data_source_secrets(self, source_definition: dict) -> list[str]:
        """Creates Databricks secrets for credentials and connection strings in a given data source definition.
        :parameter source_definition: Data source definition as a ``dict``
        :return: Created secret keys as a ``list[str]``
        """
        secret_keys = []
        source_service_name = source_definition.get("service_name")
        source_service_type = source_definition.get("type")
        if self.workspace_client is None:
            raise ValueError("workspace_client is not initialized")
        scopes = [scope.name for scope in self.workspace_client.secrets.list_scopes()]
        if "wkmigrate_credentials_scope" not in scopes:
            self.workspace_client.secrets.create_scope(scope="wkmigrate_credentials_scope")
        if source_service_type is None:
            raise ValueError("Source service type cannot be None")
        for secret in secrets[source_service_type]:
            secret_value = source_definition.get(secret)
            if secret_value is None:
                secret_value = input(f"Enter {secret} for dataset {source_service_name}")
            if self.workspace_client is None:
                raise ValueError("workspace_client is not initialized")
            self.workspace_client.secrets.put_secret(
                scope="wkmigrate_credentials_scope",
                key=f"{source_service_name}_{secret}",
                string_value=secret_value,
            )
            secret_keys.append(f"{source_service_name}_{secret}")
        return secret_keys

    def _create_notebook_task_dependencies(self, task: dict) -> str:
        """Creates a Databricks notebook if it does not exist in the target workspace.
        :parameter task: Task definition as a ``dict``
        :return: Notebook path in the target workspace as a ``str``
        """
        notebook_task = task.get("notebook_task")
        if notebook_task is None:
            raise ValueError('No "notebook_task" found in task')
        notebook_path_value = notebook_task.get('notebook_path')
        if notebook_path_value is None:
            raise ValueError('No "notebook_path" found in notebook_task')
        notebook_path = f"/Workspace{notebook_path_value}"
        cluster_definition = task.get("new_cluster")
        if cluster_definition is None:
            warnings.warn('No "new_cluster" found in task, using serverless compute')
        else:
            host_name = cluster_definition.pop("host_name")
        if task.get("unsupported"):
            return notebook_path
        try:
            if self.workspace_client is None:
                raise ValueError("workspace_client is not initialized")
            self.workspace_client.workspace.get_status(path=notebook_path)
            return notebook_path
        except Exception:
            pat = input("Notebook not found in target workspace for notebook task. Enter source workspace PAT")
            self._upload_notebook(host_name, pat, notebook_path)
            return notebook_path

    def _upload_notebook(self, host_name: str, pat: str, notebook_path: str) -> None:
        """Uploads a Databricks notebook to the target workspace.
        :parameter host_name: Workspace host as a ``str``
        :parameter pat: Workspace PAT as a ``str``
        :parameter notebook_path: Target notebook path as a ``str``
        """
        source_client = WorkspaceManagementClient(host_name=host_name, pat=pat, authentication_type="pat")
        target_folder = "/".join(notebook_path.split("/")[:-1])
        if self.workspace_client is None:
            raise ValueError("workspace_client is not initialized")
        self.workspace_client.workspace.mkdirs(target_folder)
        if source_client.workspace_client is None:
            raise ValueError("source workspace_client is not initialized")
        language = source_client.workspace_client.workspace.get_status(path=notebook_path).language
        with source_client.workspace_client.workspace.download(path=notebook_path, format=ExportFormat.SOURCE) as file:
            self.workspace_client.workspace.upload(
                path=notebook_path,
                content=file,
                format=ImportFormat.SOURCE,
                language=language,
            )

    def _create_for_each_task(self, task: dict) -> Task:
        """Creates a Databricks workflow ``Task`` object from a foreach task definition.
        :parameter task: Workflow for each task definition as a ``dict``
        :return: Workflow ``Task`` object
        """
        if "for_each_task" not in task:
            raise ValueError('No "for_each_task" value for for each task type')
        for_each_task = task.get("for_each_task")
        if for_each_task is None:
            raise ValueError('"for_each_task" cannot be None')
        inner_tasks = for_each_task.get("task")
        print(inner_tasks[0])
        if len(inner_tasks) == 1:
            task["for_each_task"]["task"] = inner_tasks[0]
        else:
            task_key = f"{task.get('task_key')}_inner_tasks"
            job_id = self.create_workflow(
                {
                    "settings": {
                        "name": task_key,
                        "tasks": inner_tasks,
                        "tags": {"CREATED_BY_WKMIGRATE": ""},
                    }
                }
            )
            task["for_each_task"]["task"] = {
                "task_key": task_key,
                "run_job_task": {"job_id": job_id},
            }
        return Task.from_dict(task)

    @staticmethod
    def _get_schedule(schedule: dict | None) -> CronSchedule | None:
        if schedule is None:
            return None
        return CronSchedule.from_dict(schedule)


@dataclass
class WorkspaceTestClient(DatabricksWorkspaceClient):
    """A mock client implementing methods to list, describe, and update objects in a Databricks workspace."""

    test_json_path: str = wkmigrate.JSON_PATH

    def get_workflow(self, job_id: int | None = None, job_name: str | None = None) -> Job:
        """Gets a workflow with the specified ID or name as a ``Job`` object.
        :parameter job_id: Job ID for the specified workflow
        :parameter job_name: Job name for the specified workflow
        :return: Workflow definition as a ``Job``
        """
        # Check the input parameters:
        if job_id is None and job_name is None:
            raise ValueError('Must provide a value for "job_id" or "job_name".')

        # Open the test workflows file:
        with open(f"{self.test_json_path}/test_workflows.json", "r") as file:
            # Load the data from JSON:
            workflows = json.load(file)

        # If a workflow ID is specified, get the workflow by ID:
        if job_id is not None:
            workflows = [workflow for workflow in workflows if workflow.get("job_id") == job_id]
            if len(workflows) == 1:
                return Job.from_dict(workflows[0])
            raise ValueError(f'No workflow found with job ID {job_id}.')

        # Otherwise, list the workflows by name:
        if job_name is not None:
            workflows = [workflow for workflow in workflows if workflow["settings"]["name"] == job_name]
            if len(workflows) == 1:
                return Job.from_dict(workflows[0])
            # If no workflow was found:
            raise ValueError(f'No workflow found with job name {job_name}.')

        # This should never be reached due to the initial check, but mypy needs it
        raise ValueError('Must provide a value for "job_id" or "job_name".')

    def create_workflow(self, job_definition: dict, translation_options: dict | None = None) -> None:
        pass
